//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_ghost.C,v 1.4 2004/08/20 22:40:18 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prom_ghost.C
//
#include "prom_grid.hh"
 
// not thread safe -- make these Promgrid variables!!!
static PromList<PromNode*> *extraGhosts = NULL;
static PromList<int> *sends = NULL;

#undef __FUNCT__   
#define __FUNCT__ "PromGrid::addGhost" 
/* PromGrid::addGhost ***************************************************
 *
 *  - find a ghost node with an ID < otherREALid, <locid,myproc> >
 *
 *  INPUT:
 *    - curr: ghost node to find (out).
 *    - proc: processor of real node.
 *    - globalID: global id of real node (key)
 *    - locID: local index of the real node
 *
 *  SIDE EFFECTS:
 *    - creates some ghosts and adds them to 'extraGhosts' and 'ghosts', 
 *      and store sends for dependancy message(#tag) < localid, newid >
 *    - "Mark"s new ghosts
 *
 *  RETURN:
 *    - was a new ghost
 */
int PromGrid::addGhost( PromNode *&curr, const int proc, const int globalID, 
			const int locID, const int tag )
{ 
  curr = NULL;                 assert(proc!=comm_.mype());  assert(proc>=0);
  
  // make new ghost - send back depency message
  curr = ghosts_.find( globalID + 1 );
  if( curr != NULL )  return 0; // got it

  if( extraGhosts == NULL ){
    extraGhosts = new PromList<PromNode*>(256); CHKNEW( extraGhosts, 1 ); 
    sends = new PromList<int>(3*256);
  }
  curr = newNode( nNodes_ + extraGhosts->getCount() ); CHKNEW( curr, 1 ); 
  curr->setType( PROM_SOLID ); // for now
  assert(curr->index_() == nLocalNd_ + ghosts_.getCount()); 
  // add to table for fast lookup (key is one based)
  bool add = ghosts_.add(curr, globalID+1); assert( add );    
  extraGhosts->addTail( curr ); // array like syntax
  curr->mark();           // mark new ghosts as do not have: coord, type
  curr->kill( FALSE );

  // set this new data
  curr->index_.setGlobal( globalID );
  curr->symbData_->ghostID.setProc( proc );
  if( locID < 0 ) SETERRQ1(1,"lid=%d",locID);
  curr->symbData_->ghostID.setLocal( locID ); assert( curr->isGhost() );
  PromList<PromGlobalID*> *list = new PromList<PromGlobalID*>(3);
  curr->symbData_->dependents = list; CHKNEW(list, 1 );

  // send master a message
  if( tag >= 0 ) {
    sends->addTail( locID+1 );
    sends->addTail( curr->index_()+1 );
    sends->addTail( proc+1 );
  }
  else assert(0);

  return 1;  
}  

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::ReceiveAddGhost" 
/* PromGrid::ReceiveAddGhost ******************************
 *
 *  - receive addGhost() messages, and resolve data structures
 *
 *  INPUT
 *    - todo: number of messages to process
 *    - rtag: tag to use 
 *    - procTable: procs that I sent to and will recv (in/out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */ 
int PromGrid::ReceiveAddGhost( int todo, const int rtag, 
			       const PromTable<int> &procTable )
{ 
  int       ierr,tt,nmess,messSz,ii,receive,proc;
  const int srtag1=getNewTag(),srtag2=getNewTag(),srtag3=getNewTag();
  MPI_Status status;
  TablePos  tpos;
  PromTable<PromList<int>*> proc_list(57); 
  PromList<int> *list;
  PromList<int*> sendlist1(50), sendlist2(50);           assert(isActive());
  PromList<MPI_Request*> smessages; 
  const int reqsz = sizeof(MPI_Request)/sizeof(int);

  if( extraGhosts ) {
    assert(extraGhosts->getCount() > 0);
    ii = nNodes_; tt = extraGhosts->getCount();
    ierr = AddNNodePointers( tt ); CHKERRQ(ierr); assert(nNodes_ == ii+tt);
    do nodes_[ii] = extraGhosts->removeHead(); while(++ii<nNodes_);
    assert(nNodes_ == ii); 
    // collect send 
    while( !sends->isEmpty() ) {
      int locID = sends->removeHead() - 1;
      int id = sends->removeHead() - 1;
      proc = sends->removeHead() - 1;
      if( !(list=proc_list.find(proc+1)) ) {
	list = new PromList<int>(64);
	proc_list.add( list, proc+1 );
      }
      list->addTail( locID+1 );
      list->addTail( id+1 );
    }
    // send 
    while( !proc_list.isEmpty() ) {
      list = proc_list.removeHead( &proc ); proc--;
      tt = list->getCount(); assert(tt%2==0 && tt>0);
      MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
      CHKERRQ(ierr);
      smessages.addTail( sreq );
      int *buff = (int*)(sreq+1), *pb = buff;
      while( !list->isEmpty() ) *pb++ = list->removeHead()-1;
      assert(pb-buff == tt);
      // send(rtag) *< otherREALid, locid >
      MPI_Isend( buff, tt, MPI_INT, proc/factor_, rtag, gComm_, sreq );
      delete list;
    }
  }
  // send capper
  tpos = procTable.getHeadPosition(); //assert(tpos!=0); 
  while( tpos ) {
    int proc = procTable.getNext( tpos ) - 1;
    MPI_Send( NULL, 0, MPI_INT, proc/factor_, rtag, gComm_ );
  }
  // recieve(rtag)
  int procstodo = procTable.getCount();
  while( procstodo ) {
    // recv(rtag) request for data (new dep)
    do{
      MPI_Iprobe( MPI_ANY_SOURCE, rtag, gComm_, &receive, &status);
      if( receive ) {
	MPI_Get_count( &status, MPI_INT, &tt ); 
	if( tt == 0 ) { // done
	  MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, rtag, gComm_, &status);
	  procstodo--; assert(procstodo >= 0);
	}
	else { // new deps
	  // receive(rtag) dependency message *< locid, otherid > 
	  ierr = ReceiveDep( rtag, tt, status.MPI_SOURCE,sendlist1, sendlist2);
	  CHKERRQ(ierr);
	} 
      }
    }while( receive );
  }
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  } 
  // send(srtag1) - new ghost
  const int np_max = procTable.getCount();
  int **sbuffs = NULL, *procs = NULL, *sizes = NULL;
  MPI_Request *sreqs = NULL;
  if( np_max ) {
    ierr = PetscMalloc( sizeof(int*) * np_max,&sbuffs);    CHKERRQ(ierr);
    ierr = PetscMalloc( sizeof(int) * np_max,&procs);      CHKERRQ(ierr); 
    ierr = PetscMalloc( sizeof(int) * np_max,&sizes);      CHKERRQ(ierr);
    ierr = PetscMalloc(sizeof(MPI_Request)*np_max,&sreqs); CHKERRQ(ierr);
  }
  int nprocs = 0;
  tpos = procTable.getHeadPosition(); //assert(tpos!=0); 
  while( tpos ) { 
    const int procj = procTable.getNext( tpos ) - 1; 
    nmess = 0; messSz = 0;
    // scan list2 for each proc
    Position pos = sendlist2.getHeadPosition(); 
    while( pos ) { 
      int *arr = sendlist2.getNext(pos); 
      if(arr[0] != procj) continue;
      nmess++; 
      messSz += arr[1]; 
    }
    int *buff;
    tt = messSz + ((sizeof(gfloat) > sizeof(int)) ? 2 : 1); 
    ierr = PetscMalloc( tt * sizeof(int),&buff );      CHKERRQ(ierr);
    messSz = 0;  buff[messSz++] = nmess; 
    if( sizeof(gfloat) > sizeof(int) ) buff[messSz++] = -1;
    pos = sendlist2.getHeadPosition(); 
    while( pos ) {  
      int *arr = sendlist2.getNext(pos); 
      if(arr[0] != procj) continue; 
      int t = arr[1];  arr[1] = -3;
      PetscMemcpy( &buff[messSz], &arr[2], t*sizeof(int) );	   
      messSz += t;
    }
    assert(tt==messSz);
    // send(srtag1)  < nmess, [-1,] nmes * < data > >
    MPI_Isend(buff,messSz,MPI_INT,procj/factor_,srtag1,gComm_,&sreqs[nprocs]);
    sbuffs[nprocs++] = buff;
  }
  assert(nprocs==np_max);
  // free send buffers
  while( !sendlist2.isEmpty() ) { 
    int *arr = sendlist2.removeHead(); assert(arr[1] == -3);
    PetscFree( arr );
  }
  // recv(srtag1) - new ghost data
  tpos = procTable.getHeadPosition(); 
  while( tpos ) { 
    procTable.getNext( tpos ); // counter only
    MPI_Probe( MPI_ANY_SOURCE, srtag1, gComm_, &status );
    const int procj = status.MPI_SOURCE * factor_; 
    MPI_Get_count( &status, MPI_INT, &messSz );
    assert(procj != comm_.mype()); 
    int *buff; ierr = PetscMalloc(messSz*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(srtag1) nmess, [-1,] nmes * < data > 
    MPI_Recv( buff, messSz, MPI_INT, procj/factor_, srtag1, gComm_, &status);
    nmess = *pb++; 
    if( sizeof(gfloat) > sizeof(int) ){ tt = *pb++; assert(tt == -1); }
    while( nmess-- ){
      todo--;       
      ierr = ReceiveGhost( procj, pb );                       CHKERRQ(ierr);
      if((size_t)pb%sizeof(gfloat)!=0){tt = *pb++; assert(tt==-2);}; 
    }
    assert( pb - buff == messSz );
    MPI_Get_count( &status, MPI_INT, &tt ); assert(pb-buff == tt);
    ierr = PetscFree(buff); buff = NULL;                     CHKERRQ(ierr);
  }
  assert(todo==0);
  // wait and free send buffers
  if( nprocs ) {
    MPI_Status *statuses;
    ierr = PetscMalloc(nprocs*sizeof(MPI_Status),&statuses); CHKERRQ(ierr);
    MPI_Waitall( nprocs, sreqs, statuses );   
    for( tt = 0 ; tt < nprocs ; tt++ ) {     
      if( sbuffs[tt] != NULL ){ ierr = PetscFree( sbuffs[tt] );  }
      sbuffs[tt] = NULL; 
    } 
    PetscFree( sbuffs );      PetscFree( procs );    PetscFree( sizes ); 
    PetscFree( statuses );    PetscFree( sreqs );
  }
   
  // send new dep messages (srtag2)
  Position pos = sendlist1.getHeadPosition(); 
  while( pos ) { 
    int *buff = sendlist1.getNext( pos );
    int proc = buff[0];
    if( (list=proc_list.find(proc+1)) == NULL ) {
      list = new PromList<int>(52); CHKNEW(list,1);
      proc_list.add( list, proc+1 );
    }
    list->addTail( buff[1]+1 ); list->addTail( buff[2]+1 ); 
    list->addTail( buff[3]+1 );
    PetscFree(buff);                               
    todo++;
  }
  // isend
  int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 
  PetscMemzero( sbuf, comm_.np()*sizeof(int)/factor_ ); // collect mess. counts
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    tt = list->getCount(); assert(tt%3==0 && tt>0);
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)(sreq+1), *pb = buff;
    while( !list->isEmpty() ) { ii = list->removeHead()-1; *pb++ = ii; }
    assert(pb-buff == tt);
    // send(srtag2) *< gid, locid, proc >
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, srtag2, gComm_, sreq );
    delete list;
    sbuf[proc/factor_] = 1;
  }
  // global!!! -- need as have little idea of who is sending stuff
  for( tt = 0 ; tt < comm_.np()/factor_ ; tt++ ) rbuf[tt] = 1;
  //MPI_Reduce_scatter( sbuf, &tt, rbuf, MPI_INT, MPI_SUM, gComm_ );
  MPI_Allreduce( sbuf, rbuf, comm_.np()/factor_, MPI_INT, MPI_SUM, gComm_);  
  tt = rbuf[comm_.mype()/factor_]; // num messages to recieve
  // processes extra deps
  ierr = CollectNewDeps( todo, tt, srtag2, srtag3 );  CHKERRQ(ierr);
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  } 
  if( extraGhosts ){
    delete extraGhosts; extraGhosts = NULL; 
    delete sends; sends = NULL; 
  }
  // debug
  MPI_Iprobe( MPI_ANY_SOURCE, rtag, gComm_, &tt, &status ); assert(!tt);
  MPI_Iprobe( MPI_ANY_SOURCE, srtag1, gComm_, &tt, &status ); assert(!tt);
  MPI_Iprobe( MPI_ANY_SOURCE, srtag3, gComm_, &tt,&status ); assert(!tt);
  MPI_Iprobe( MPI_ANY_SOURCE, srtag2, gComm_, &tt, &status ); assert(!tt);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::ReceiveDep" 
/* PromGrid::ReceiveDep *********************************************
 *
 *  - receive a new dependent for a bound
 *  - matches addGhost()
 *
 *  INPUT:
 *   - rtag: MPI tag to receive
 *   - sendlist1: list of < proc2 <gid, locidj, procj> > (in/out)
 *   - sendlist2: list of < data > (in/out)
 *
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *    - Petsc error code
 */
int PromGrid::ReceiveDep( const int rtag, const int sz, const int part,
			  PromList<int*> &sendlist1, PromList<int*> &sendlist2)
{
  PromNode   *boun;                                          assert(sz>0);
  int         procj, locidj, locidi, ierr, flag, proc, *buff, *pb;
  MPI_Status  status; 
  PromGlobalID *dep2;

  ierr = PetscMalloc( sz * sizeof(int), &buff );                 CHKERRQ(ierr);
  // receive(rtag) *< locid, otherlocid > -- matches addGhost()
  MPI_Recv( buff, sz, MPI_INT, part, rtag, gComm_, &status );        
  pb = buff;
  while( pb-buff < sz ) {
    PromContext::Read( locidi, pb ); PromContext::Read( locidj, pb );
    boun = nodes_[locidi];     assert( locidi >= 0 && locidi < nLocalNd_ );
    procj =  part * factor_;
    // add a dep - send MPI(stag1) <coord,...> message in  sendlist2
    ierr = AddDep( boun, procj, locidj, sendlist2 );            CHKERRQ(ierr);
    // send other 
    Position pos = boun->symbData_->dependents->getHeadPosition();
    assert(pos!=NULL); 
    flag = 0;
    while( pos ) {
      dep2 = boun->symbData_->dependents->getNext( pos ); 
      if( dep2->proc() != procj ) {
	proc = dep2->proc(); 
	// send(tag2) 'proc2' dep message < proc, gid, locidj, procj >
	int *pb2; ierr = PetscMalloc(4*sizeof(int),&pb2); CHKERRQ(ierr);
	int *pb3 = pb2; 
	sendlist1.addTail( pb3 ); 
	PromContext::Write( proc, pb3 );
	PromContext::Write( boun->index_.global(), pb3 ); 
	PromContext::Write( locidj, pb3 );
	PromContext::Write( procj, pb3 );
      }
      else flag++;
    }
    assert(flag==1);
  }
  assert(pb-buff == sz);  PetscFree(buff);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::AddDep" 
/* PromGrid::AddDep ***************************************************
 *
 *  - find a ghost node with an ID < id, proc  >
 *
 *  INPUT: 
 *    - boun: node to check
 *    - procj: processor of real node.
 *    - locidj: local ID of ghost on 'proc'
 *    - sendlist2: list of < data > (in/out)
 *
 *  SIDE EFFECTS:
 *    - creates some dependents and adds them to 'boun->dependents', and sends 
 *      dependancy message(#tag) < locid, gid, coord, type, elem >
 *
 *  RETURN:
 *    - Petsc error code
 */
int PromGrid::AddDep( PromNode *boun, const int procj, const int locidj, 
		      PromList<int*> &sendlist2 ) 
{
  int          *pb2, nadj, tt, ierr; 
  Position      pos;
  PromGlobalID *dep,*dep2;
  int bf[PROM_BSZ+1], *sbuf = (size_t)bf%sizeof(shpfloat) == 0 ? bf : bf+1; 

  if( !boun->symbData_->dependents ) {
    boun->symbData_->dependents = new PromList<PromGlobalID*>(3);
    CHKNEW( boun->symbData_->dependents, 1 );
  }
  assert( boun->isBound() );
  
  dep = new PromGlobalID( locidj, procj ); CHKNEW( dep, 1 ); assert(locidj>=0);
  boun->symbData_->dependents->addTail( dep );
  // send dependent ghost data
  int *pb = sbuf;  
 
  // <lid,gid,node,nadj,nadj*<gid>,ndep,ndep*<loc,proc>>
  PromContext::Write( locidj, pb ); 
  PromContext::Write( boun->index_.global(), pb );
  ierr = boun->Write( &pb ); CHKERRQ(ierr);
  pb2 = pb; *pb++ = -4;
  TablePos tpos = boun->symbData_->adjac.getHeadPosition(); nadj = 0;
  while( tpos ) { 
    PromNode *n2 = boun->symbData_->adjac.getNext( tpos ); 
    if( n2->isExtended() ){ nadj++; PromContext::Write(n2->index_.global(),pb);}
  }
  assert(*pb2==-4); *pb2 = nadj; 
  // deps
  *pb++ = boun->symbData_->dependents->getCount();
  pos = boun->symbData_->dependents->getHeadPosition(); assert(pos!=NULL); 
  while( pos ) {
    dep2 = boun->symbData_->dependents->getNext( pos ); 
    tt = dep2->local();
    PromContext::Write( tt, pb ); PromContext::Write( dep2->proc(), pb ); 
  }
  if( (size_t)pb%sizeof(gfloat) != 0) *pb++ = -2; // make even
  if( pb-sbuf >= PROM_BSZ )  SETERRQ(1,"buffer to small");
  // send(tag1) ghost a dep message 
  int sz = pb - sbuf; 
  int *buff; ierr = PetscMalloc((sz+2)*sizeof(int),&buff ); CHKERRQ(ierr);
  buff[0] = procj; buff[1] = sz;
  PetscMemcpy( &buff[2], sbuf, sz*sizeof(int) ); 
  sendlist2.addTail( buff );
  
  return 0;
}
   
#undef __FUNCT__  
#define __FUNCT__ "PromGrid::ReceiveGhost" 
/* PromGrid::ReceiveGhost *********************************************
 *
 *  - receive a new ghost data, can send "stag"
 *  - matches AddDep(), called from ReceiveDep()
 *
 *  INPUT:
 *   - proc: 
 *   - pb: buffer re read
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::ReceiveGhost( const int source_proc, int *&pb )
{
  int         ierr,gidi,locidi,ii,gid,nadj,id,flag,proc2, master=masterProc();
  PromNode    *n1,*ghos;
  
  // recive(rtag) ghost data 
  // <lid,gid,<node>,nadj,nadj*<gid>,ndeps*<local,proc> >
  assert((size_t)pb%sizeof(gfloat)==0);     
  PromContext::Read( locidi, pb ); PromContext::Read( gidi, pb ); 
  id = gidi - proc_gnode_[source_proc];
  ghos = ghosts_.find(gidi+1); 
  assert(ghos!=NULL && ghos->index_.global() == gidi); 
  assert( ghos->isMarked() );    int olsst = (int)ghos->getState();
  ierr = ghos->Read( parBank_, &pb ); CHKERRQ(ierr);
  ghos->mark(FALSE);
  assert(ghos->index_() == locidi && ghos->index_.global() == gidi);
  if( (int)ghos->getState() != olsst && olsst != PROM_UNKNOWN ){
    PetscPrintf(PETSC_COMM_SELF,"[%d]%s ERROR %d oldst=%s,newst=%s\n",
		comm_.mype(),__FUNCT__,ghos->index_.global(),
		olsst==PROM_DELETED?"DELETED":
		olsst==PROM_SELECTED?"SELECTED":"UNKNOWN",
		ghos->getState()==PROM_DELETED?"DELETED":
		ghos->getState()==PROM_SELECTED?"SELECTED":"UNKNOWN");
    assert(ghos->getState()==PROM_DELETED);
    assert(ghos->symbData_->parent==NULL);
    assert(ghos->isChildListEmpty());
    SETERRQ(1,"oldst=SELECTED,newst=DELETED");
  }

  // set type...
  PromContext::Read( nadj, pb );        assert( ghos->isGhost() );
  while( nadj-- ) {
    PromContext::Read( gid, pb ); 
    id = gid - proc_gnode_[master];
    if( id < 0 || id >= nLocalNd_ ) n1 = ghosts_.find( gid + 1 );
    else n1 = nodes_[id];
    if( n1 ) {
      assert( n1 != ghos );
      if( ghos->symbData_->adjac.add(n1) ) {
	n1->symbData_->adjac.add(ghos);
	assert( ghos->isBound() || ghos->isGhost() );
      }
    }
  }
  // deps
  assert(ghos->symbData_->dependents&&ghos->symbData_->dependents->isEmpty());
  PromContext::Read( nadj, pb ); assert( nadj > 0 ); PromGlobalID *dep; 
  for( ii = flag = 0 ; ii < nadj ; ii++ ) {
    PromContext::Read( id, pb ); PromContext::Read( proc2, pb );
    if( proc2 == comm_.mype() ) { // ref to real node (copy of realID)
      assert( id == ghos->index_() ); flag++; 
      proc2 = source_proc; id = ghos->symbData_->ghostID.local();
      if(id>=nglobal())SETERRQ2(1,"id(%d) >= nglobal(%d)",id,nglobal());
    }
    // ref to other ghost 
    dep = new PromGlobalID( id, proc2 );  CHKNEW( dep, 1 );
    ghos->symbData_->dependents->addHead( dep );
  }
  ghos->setExterior(); // for geo
      
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::CollectNewDeps" 
/* PromGrid::CollectNewDeps *********************************************
 *
 *  -  Collect new deps for ghosts
 *
 *  INPUT:
 *   - todo: 
 *   - rtag: MPI rtag to receive
 *   - srtag1: MPI tag to send/receive
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGrid::CollectNewDeps(int todo, int nrecv, const int rtag2,
			     const int srtag3 )
{
  int ierr,receive,tt;
  MPI_Status status;

  while( nrecv || todo ){
    // recv(rtag2) - new dep
    do{ 
      MPI_Iprobe( MPI_ANY_SOURCE, rtag2, gComm_, &receive, &status );
      if( receive ) {
	MPI_Get_count( &status, MPI_INT, &tt ); 
	ierr = ReceiveGhostDep( rtag2, srtag3, status.MPI_SOURCE, tt );
	CHKERRQ(ierr);
	nrecv--;       assert(nrecv >= 0);
      }
    }while( receive );
    // recv(srtag3) - response
    do{ 
      MPI_Iprobe( MPI_ANY_SOURCE, srtag3, gComm_, &receive, &status );
      if( receive ) {
	MPI_Get_count( &status, MPI_INT, &tt ); assert(tt==1);
	// receive(srtag3) 
	MPI_Recv( &tt, 1, MPI_INT, status.MPI_SOURCE, srtag3, gComm_, &status);
	todo -= tt;  assert( todo >= 0 ); 
      }
    }while( receive );
  }
  assert(todo==0);
 
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::ReceiveGhostDep" 
/* PromGrid::ReceiveGhostDep *********************************************
 *
 *  -  Collect new deps for ghosts
 *
 *  INPUT:
 *   - rtag2: MPI tag to recv
 *   - stag3: MPI tag to send
 *   - part:
 *   - sz:
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGrid::ReceiveGhostDep( const int rtag2, const int stag3, 
			       const int part, const int sz )
{
  MPI_Status status;
  int ierr,id,proc,gid,tt=0,*buff,*pb; 

  ierr = PetscMalloc(sz*sizeof(int),&buff); pb = buff; CHKERRQ(ierr);
  
  // receive(rtag2) dependency message < gid, locid, proc >
  MPI_Recv( buff, sz, MPI_INT, part, rtag2, gComm_, &status );
  MPI_Get_count( &status, MPI_INT, &id );   assert(id == sz);
  while(pb-buff < sz) {
    PromContext::Read( gid, pb ); PromContext::Read( id, pb ); 
    PromContext::Read( proc, pb );
    //
    PromNode *ghos = ghosts_.find( gid+1 ); assert(ghos!=NULL);
    PromGlobalID *dep = new PromGlobalID( id, proc ); CHKNEW( dep, 1 );
    ghos->symbData_->dependents->addHead( dep ); 
    tt++;
  }
  assert(pb-buff == sz);  PetscFree(buff);

  // respond w/ number of recvs
  MPI_Send( &tt, 1, MPI_INT, part, stag3, gComm_ );

  return 0;
}
