//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_kkt.C,v 1.7 2004/08/20 07:19:03 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prom_kkt.C
//   
#include "prometheus.hh"

// lapack routines (FORTRAN)
#if defined(__cplusplus)
extern "C"{
#endif
  void *dgetrf_(const int*, const int*, double*, const int*, int*, int*);
  void *dgetrs_(const char*, const int*, const int*, const double*, const int*,
		const int*, double*, const int*, int*);
#if defined(__cplusplus)
}
#  endif

/******************************************************************************
 *  PromPC_base
 ************************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PromPC_base::CreateKKTSmoother"
/* PromPC_base::CreateKKTSmoother ****************************************
 *
 *  - create new KKT ASM PC object (non-top or one level solve)
 * 
 *  INPUT: 
 *    - CC:
 */     
int PromPC_base::CreateKKTSmoother( const PromMatrix *const AA, 
				    const PromMatrix * const CC,
				    const PromCRVector *const XX,
				    const int naggs, PromIS *CRIS, 
				    const PromMatrix*const CCt ) 
{ 
  int ierr; bool fg; char st[32]; assert( this != NULL ); 

  PetscFunctionBegin;  
  if( kkt_ != NULL ) delete kkt_;  
  // factory
  bool ilu = FALSE;
  ierr = options_.GetOptionString("-cr_pc_type",st,32,&fg);
  if( fg && strcmp(st,"segregated" )==0 ) {
    assert( CCt != NULL ); assert(naggs == 0);
    kkt_ = new PromPCKKT_Seg( CC, CCt ); 
  }
  else {
    if( fg && strcmp(st,"ilu")==0 ) {
      kkt_ = new PromPCKKT_Shell( AA, CC );  // sets operator
      ilu = TRUE; assert(naggs==1);
      PromISDestroy( CRIS[0] ); PetscFree( CRIS ); 
    }
    else {
      if( naggs == 0 ) SETERRQ(1,"naggs == 0");
      kkt_ = new PromPCKKT_ASM( CC, naggs, CRIS );  // sets operator
    }
  }

  // create PCs -- creates 'pc_'
  ierr = kkt_->Create( options_, perf_mon_, AA, XX, ilu ); CHKERRQ(ierr);
  if( setup_called_ != 2 ) {
    setup_called_ = 2; // now setup -- just line 'PromPC::CreateASM'  
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PromPC_base::CreateKKTSolver"
/* PromPC_base::CreateKKTSolver ****************************************
 *
 *  - create new KKT PC object, whole local matrix (processor block) LU PC
 * 
 *  INPUT: 
 *    - AA:
 */     
int PromPC_base::CreateKKTSolver(const PromMatrix *const AA,
				 const PromMatrix *const CC,
				 const PromCRVector *const XX ) 
{
  int ierr; assert( this != NULL ); 
  PetscFunctionBegin;
  if( kkt_ != NULL ) delete kkt_;
  kkt_ = new PromPCKKT_Shell( AA, CC ); // sets operator, 'CRIS_'
  ierr = kkt_->Create( options_, perf_mon_, AA, XX ); CHKERRQ(ierr); // 'ASM'
  if( setup_called_ != 2 ) {
    setup_called_ = 2; // now setup -- just line 'PromPC::CreateASM'
  }
  PetscFunctionReturn(0);
}

/************************************************************************
 *  PromPCKKT
 ************************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PromPCKKT::~PromPCKKT"
PromPCKKT::~PromPCKKT()
{
  if( r_ != NULL ) delete r_; // 
  if( work_ != NULL ) delete work_; // 
  if( dpc_ != NULL ) delete dpc_; 
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT::SetUp"
/* PromPCKKT::SetUp ****************************************
 *
 *  - subdomain factors
 * 
 *  INPUT: 
 *   - AA: primal part
 */     
int PromPCKKT::SetUp() 
{
  int ierr; assert( this != NULL ); assert( dpc_ != NULL ); 
#if defined(PROM_TRACE_PC_SETUP)
  PetscPrintf(PETSC_COMM_WORLD,"\t[%d]%s this=%p\n",0,__FUNCT__,this);
#endif
  // put 'aux_mats_' in osm
  ierr = dpc_->SetUp(); CHKERRQ(ierr); // get submatrices

  return 0;
}

/************************************************************************
 *  PromPCKKT_ASM
 ************************************************************************/

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_ASM::PromPCKKT_ASM"
/* PromPCKKT_ASM::PromPCKKT_ASM ****************************************
 *
 *  INPUT: 
 *   - CC: 
 *   - naggs: 
 *   - CRIS: 
 */     
PromPCKKT_ASM::PromPCKKT_ASM( const PromMatrix*const CC,const int naggs,
			      PromIS *CRIS) : PromPCKKT(CC),
#if defined(PROM_USE_PETSC)
  aux_mats_(NULL), d_scatters_(NULL), 
#else // petra
#endif
  PrimIS_(NULL), CRIS_(CRIS), mask_(NULL), zeros_(NULL),
  nBlocks_(naggs), nMask_(0), lid_isInKKT_(NULL)
{}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_ASM::getNumBlocks"
/* PromPCKKT_ASM::getNumBlocks ****************************************
 *
 */     
int PromPCKKT_ASM::getNumBlocks() const
{
  assert(this!=NULL);
  return dpc_->getNumBlocks();
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_ASM::~PromPCKKT_ASM"
/* PromPCKKT_ASM::~PromPCKKT_ASM ****************************************
 *
 *  INPUT: 
 */     
PromPCKKT_ASM::~PromPCKKT_ASM()   
{
  int ii;
#if defined(PROM_USE_PETSC)
  VecScatter *scats = (VecScatter*)d_scatters_;
  Mat *mats = (Mat*)aux_mats_;
#else // petra
#endif
  int mype;  MPI_Comm_rank( PETSC_COMM_WORLD, &mype );
  assert(CRIS_ != NULL); assert(PrimIS_ != NULL); assert(nBlocks_>0);
  for( ii = 0 ; ii < nBlocks_ ; ii++ ) {
    PromISDestroy( CRIS_[ii] ); 
    PromISDestroy( PrimIS_[ii] ); 
#if defined(PROM_USE_PETSC)
    MatDestroy(mats[ii]);
#else // petra
#endif
  }
  assert(nBlocks_ > 0);
  PetscFree( CRIS_ ); PetscFree( PrimIS_ );
#if defined(PROM_USE_PETSC)
  PetscFree(mats);
#else // petra
#endif
  assert( dpc_ != NULL );
#if defined(PROM_USE_PETSC)
  for( ii = 0 ; ii < nBlocks_ ; ii++ ) {
    VecScatterDestroy(scats[ii]); 
  }
  PetscFree(scats); 
#else // petra
#endif 
  delete dpc_; dpc_ = NULL;
  if( lid_isInKKT_ != NULL ) delete lid_isInKKT_; 
  if(zeros_ != NULL)PetscFree(zeros_);
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_ASM::CreatePrimAggs"
/* PromPCKKT_ASM::CreatePrimAggs ****************************************
 *
 *  - create 'PrimIS_' with 'CRIS_'
 * 
 *  INPUT: 
 *   - AA: primal part
 */     
int PromPCKKT_ASM::CreatePrimAggs( const PromMatrix *const AA, bool ilu ) 
{
  int ierr,bid,nloc=AA->map_.nlocal(); assert(AA->map_.isConstNDF()); 
  const int ndf=AA->map_.constNDFSize(), my0Nd = AA->map_.my0Nd();
  PromTable<int> geqtab(517);  assert(lid_isInKKT_==NULL);
  
  PetscFunctionBegin;
  if(!ilu) lid_isInKKT_ = new PromBitArr(nloc);
  assert( nBlocks_ > 0 && PrimIS_ == NULL);
  ierr = PetscMalloc( sizeof(PromIS)*nBlocks_, &PrimIS_ ); CHKERRQ(ierr);
  for( bid = 0 ; bid < nBlocks_ ; bid++ ) {
    // get eqs in this agg of CRs
    PromIS lis = CRIS_[bid]; int ii, nids, *gids;
    ierr = PromISGetSize( lis, &nids ); CHKERRQ(ierr);
    ierr = PromISGetIndices( lis, &gids ); CHKERRQ(ierr);
    for( ii = 0 ; ii < nids ; ii++ ){
      int lmGID = gids[ii], ncols, *colids;
      ierr = C_->GetRow(lmGID, &ncols, &colids ); CHKERRQ(ierr);
      for( int jj=0 ; jj < ncols ; /*void*/ ){
	const int gid = colids[jj]/ndf, geq0 = gid*ndf;
	for(int kk=0;kk<ndf;kk++) geqtab.add(geq0 + kk + 1);
	//
	const int lid = gid - my0Nd;
	if(lid>=0 && lid < nloc) lid_isInKKT_->SetAt(lid,TRUE);
	// inc 'jj' to next geq0
	while( jj < ncols && colids[jj] < geq0+ndf ) jj++;
      }
      ierr = C_->RestoreRow(lmGID, &ncols, &colids );CHKERRQ(ierr);
    }
    // make IS
    int sz1 = geqtab.getCount(), buffer[PROM_BSZ], *buf; 
    if(sz1>PROM_BSZ){ierr = PetscMalloc(sizeof(int)*sz1, &buf); CHKERRQ(ierr);}
    else buf = buffer;
    for( ii = 0 ; ii < sz1 ; ii++ ) buf[ii] = geqtab.removeHead() - 1;
    PromIS gis;  assert( geqtab.isEmpty() );
    ierr = PromISCreateGeneral(PETSC_COMM_SELF,sz1,buf,&gis);   CHKERRQ(ierr);
    ierr = ISSort( gis );                                       CHKERRQ(ierr);
    if( buf != buffer ) PetscFree(buf);
    //
    PrimIS_[bid] = gis;
    ierr = PromISRestoreIndices( lis, &gids ); CHKERRQ(ierr);
  }
  // get mask arry
  if( lid_isInKKT_ != NULL ) {
    const PromBitArr &barr = *lid_isInKKT_; 
    nMask_ = ndf*barr.getCount();   assert(zeros_==NULL); 
    if( nMask_ > 0 ) {
      ierr = PetscMalloc(nMask_*(sizeof(int)+sizeof(double)), &zeros_);
      CHKERRQ(ierr); mask_ = (int*)(zeros_+nMask_);
      int eq = AA->map_.my0Eq(), nc = 0;
      for( int ii = 0 ; ii < nloc; ii++, eq += ndf ){
	if( barr[ii] ){
	  for(int j=0;j<ndf;j++,nc++){ mask_[nc] = eq+j; zeros_[nc]=0.0;}
	}
      }
      assert(nc==nMask_);
    }
  }

  PetscFunctionReturn(0);
}

#if defined(PROM_USE_PETSC)
static int PromPCSetUp_ASM( PC pc );
EXTERN_C_BEGIN
EXTERN int MatOrdering_QMD(Mat,MatOrderingType,IS*,IS*);
EXTERN int MatOrdering_ND(Mat,MatOrderingType,IS*,IS*);
EXTERN int MatOrdering_RCM(Mat mat,MatOrderingType type,IS *row,IS *col);
EXTERN_C_END
EXTERN int MatAdjustForInodes(Mat,IS *,IS *);
// hack to get data, not thread safe!
static VecScatter *g_vecScatar_cr = NULL;  // (out) scatter for CR vals
static Mat *g_matarr = NULL;     // (out) KKT sub-matrix
static Mat g_C = NULL;         // (in) CRs
static PromIS *g_CRISs = NULL; // (in) CR blocks
static Vec g_cr_vec = NULL;      // (in) CR vector (map) 
static bool use_lu_subpc = FALSE;
static int g_nMats = 0;

/* hack for LU setup
   typedef struct {
   Mat             fact;         
   PetscReal       actualfill;   
   int             inplace;      
   IS              row,col;      
   MatOrderingType ordering;     
   PetscTruth      reuseordering;
   PetscTruth      reusefill;    
   MatLUInfo       info;
   } PC_LU;
   
   typedef struct {
   Mat               fact;       
   MatOrderingType   ordering;   
   IS                row,col;    
   void              *implctx;   
   PetscTruth        inplace;    
   PetscTruth        reuseordering;
   
   PetscTruth        usedt;        
   PetscTruth        reusefill;    
   double            actualfill;   
   MatILUInfo        info;
   PetscTruth        single_precision_solve; 
   } PC_ILU; */

typedef struct {
  Mat             fact;         
  PetscReal       actualfill;   
  int             inplace;      
  IS              row,col;      
  MatOrderingType ordering;     
  PetscTruth      reuseordering;
  PetscTruth      reusefill;    
  MatFactorInfo   info;
} PC_LU;

typedef struct {
  Mat               fact;           
  MatOrderingType   ordering;       
  IS                row,col;        
  void              *implctx;       
  PetscTruth        inplace;        
  PetscTruth        reuseordering;  
  
  PetscTruth        usedt;          
  PetscTruth        reusefill;      
  double            actualfill;     
  MatFactorInfo     info;
} PC_ILU;

#endif
#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_ASM::Create"
/* PromPCKKT_ASM::Create ****************************************
 *
 *  - create 'pc_'
 *
 *  INPUT: 
 *   - opts:
 *   - perf:
 *   - AA: primal part
 */     
int PromPCKKT_ASM::Create( const PromOptions &opts, 
			   PromPerfMonitor &perf,
			   const PromMatrix *const AA,
			   const PromCRVector *const XX, bool ilu /*=FALSE*/ ) 
{
  int ierr; assert(this!=NULL); assert(dpc_==NULL); 

  PetscFunctionBegin;
  // create prime ISs with 'CRIS_' 
  ierr = CreatePrimAggs( AA, ilu ); CHKERRQ(ierr); // can call 'Shell.' 

  // create 'dpc_'
  assert(dpc_ == NULL);
  dpc_ = new PromPC( opts, perf, AA->MPIComm() );
  ierr = dpc_->SetType( PROMPCASM, XX );  CHKERRQ(ierr);
  ierr = dpc_->SetOperator( AA );  CHKERRQ(ierr);
#if defined(PROM_USE_PETSC)
  PC asmpc = dpc_->pc_; 
  ierr = PCASMSetOverlap( asmpc, 0 );              CHKERRQ(ierr);
  ierr = PCASMSetType( asmpc, PC_ASM_RESTRICT );   CHKERRQ(ierr);
  ierr = PCASMSetLocalSubdomains( asmpc, nBlocks_, PrimIS_ ); CHKERRQ(ierr);
  // create 'aux_mats_' 'd_scatters_' 
  ierr = PetscMalloc( sizeof(Mat)*nBlocks_, &aux_mats_ ); CHKERRQ(ierr);

  // hack for special setup
  g_C = C_->mat_; g_matarr = (Mat*)aux_mats_; use_lu_subpc = !ilu;
  g_CRISs = CRIS_; g_cr_vec = XX->p_->vec_; g_nMats = nBlocks_;
  asmpc->ops->setup = PromPCSetUp_ASM;
  ierr = PCSetUp( asmpc ); CHKERRQ(ierr); // this will call 'PromPCSetUp_ASM'

  // out
  d_scatters_ = (void*)g_vecScatar_cr;
  g_vecScatar_cr=NULL; g_matarr=NULL; g_C=NULL; g_CRISs=NULL; g_cr_vec=NULL;
  // real setup (factors)
  ierr = PCSetUpOnBlocks( asmpc );                CHKERRQ(ierr);
  dpc_->setup_called_ = 2; // all setup -- like 'CreateASM'
#else // petra
#endif
  PetscFunctionReturn(0);
}

#if defined(PROM_USE_PETSC)
/* PromPCSetUp_ASM ****************************************
 *
 *  hacked copy of PCSetUp_ASM
 */   
#undef __FUNCT__  
#define __FUNCT__ "PromPCSetUp_ASM"
static int PromPCSetUp_ASM( PC pc )
{
  PETSc_ASM *osm = (PETSc_ASM*)pc->data;
  int      n_local,i,ierr,pm,sz,cm,max_sz,mype;
  const int n_local_true = osm->n_local_true; 
  MatReuse scall = MAT_REUSE_MATRIX;
  IS       isl;
  KSP      subksp;
  PC       subpc;
#if defined(PROM_TRACE_PC_SETUP)
  PetscPrintf(PETSC_COMM_WORLD,"\t[%d]%s pc=%p osm->n=%d\n",0,__FUNCT__,pc,osm->n);
#endif
  PetscFunctionBegin;
  MPI_Comm_rank( PETSC_COMM_WORLD, &mype );
  // hack check
  if( g_C == NULL ) SETERRQ(1,"PromPCSetUp_ASM called with global inits");
  if ( pc->setupcalled == 0 ) {
    if (osm->n == PETSC_DECIDE && osm->n_local_true == PETSC_DECIDE) { 
      assert(0); 
      /* no subdomains given, use one per processor */
    } else if (osm->n == PETSC_DECIDE) { /* determine global number of subdomains */
      int inwork[2],outwork[2];
      inwork[0] = inwork[1] = osm->n_local_true;
      MPI_Allreduce(inwork,outwork,1,MPI_2INT,PetscMaxSum_Op,pc->comm);
      osm->n_local = outwork[0];
      osm->n       = outwork[1];
    }
    n_local = osm->n_local;
    assert(n_local>0); 
    //n_local_true = osm->n_local_true;  
    assert (osm->is != NULL ); /* build the index sets */

    assert(osm->ksp == NULL);
    ierr   = PetscMalloc((n_local_true+1)*sizeof(KSP **),&osm->ksp);
    CHKERRQ(ierr);
    ierr   = PetscMalloc(n_local*sizeof(VecScatter **),&osm->scat);
    CHKERRQ(ierr);
    ierr   = PetscMalloc(2*n_local*sizeof(Vec **),&osm->x);CHKERRQ(ierr);
    osm->y = osm->x + n_local;

    /*  Extend the "overlapping" regions by a number of steps  */
    //ierr=MatIncreaseOverlap(pc->pmat,n_local_true,osm->is,osm->overlap);
    //CHKERRQ(ierr);
    for (i=0; i<n_local_true; i++) {
      ierr = ISSort( osm->is[i] ); CHKERRQ(ierr);  
    }
    /* create the local work vectors and scatter contexts */
    assert(g_vecScatar_cr==NULL);
    ierr = PetscMalloc(sizeof(VecScatter)*n_local, &g_vecScatar_cr);  
    CHKERRQ(ierr);
    for (i=max_sz=0; i<n_local_true; i++) {
      ierr = PromISGetSize(osm->is[i],&pm);CHKERRQ(ierr); 
      PromIS cris = g_CRISs[i]; // larger array (+cm)
      ierr = PromISGetSize( cris, &cm ); CHKERRQ(ierr);	
      if(cm+pm > max_sz) max_sz = cm+pm;
      ierr = VecCreateSeq(PETSC_COMM_SELF,pm+cm,&osm->x[i]);CHKERRQ(ierr);
      ierr = VecDuplicate(osm->x[i],&osm->y[i]);CHKERRQ(ierr);
      // normal primal scatter
      ierr = ISCreateStride(PETSC_COMM_SELF,pm,0,1,&isl);CHKERRQ(ierr);
      //ierr = VecScatterCreate(pc->vec,osm->is[i],osm->x[i],isl,&osm->scat[i]);
      CHKERRQ(ierr);
      ierr = ISDestroy(isl);CHKERRQ(ierr);
      // extra scatter with CRIS
      ierr = ISCreateStride(PETSC_COMM_SELF,cm,pm,1,&isl);CHKERRQ(ierr);
      ierr = VecScatterCreate(g_cr_vec,cris,osm->x[i],isl,&g_vecScatar_cr[i]);
      CHKERRQ(ierr); // global IDs ???
      ierr = ISDestroy(isl);CHKERRQ(ierr);
    }
    for (i=n_local_true; i<n_local; i++) {
      ierr = VecCreateSeq(PETSC_COMM_SELF,0,&osm->x[i]);CHKERRQ(ierr);
      ierr = VecDuplicate(osm->x[i],&osm->y[i]);CHKERRQ(ierr);
      ierr = ISCreateStride(PETSC_COMM_SELF,0,0,1,&isl);CHKERRQ(ierr);
      //ierr = VecScatterCreate(pc->vec,isl,osm->x[i],isl,&osm->scat[i]);
      CHKERRQ(ierr);
      ierr = ISDestroy(isl);CHKERRQ(ierr);
      // extra (dummy) scatter with CRIS
      ierr = ISCreateStride(PETSC_COMM_SELF,0,0,1,&isl);CHKERRQ(ierr);
      ierr = VecScatterCreate(g_cr_vec,isl,osm->x[i],isl,&g_vecScatar_cr[i]);
      CHKERRQ(ierr);
      ierr = ISDestroy(isl);CHKERRQ(ierr);
    }
    /* 
       Create the local solvers. (and exta matrices)
    */
    for (i=0; i<n_local_true; i++) {
      ierr = PromISGetSize( g_CRISs[i], &cm );   CHKERRQ(ierr);
      ierr = ISGetLocalSize( osm->is[i], &pm );  CHKERRQ(ierr); sz = pm+cm;
      //
      ierr = KSPCreate( PETSC_COMM_SELF, &subksp );   CHKERRQ(ierr);
      PetscLogObjectParent(pc,subksp);
      ierr = KSPSetType( subksp, KSPPREONLY );  CHKERRQ(ierr);
      ierr = KSPGetPC( subksp, &subpc );         CHKERRQ(ierr);
      if( use_lu_subpc ) ierr = PCSetType( subpc, PCLU );    
      else ierr = PCSetType( subpc, PCILU ); 
      CHKERRQ(ierr);
      //
      const char *prefix;
      ierr = PCGetOptionsPrefix(pc,&prefix);CHKERRQ(ierr);
      ierr = KSPSetOptionsPrefix(subksp,prefix);CHKERRQ(ierr);
      ierr = KSPAppendOptionsPrefix(subksp,"sub_");CHKERRQ(ierr);
      ierr = KSPSetFromOptions(subksp);CHKERRQ(ierr);
      osm->ksp[i] = subksp;
      // create extra (kkt) matrix
      cm = (sz<100) ? sz : (sz<1000) ? 150 : (sz<10000) ?  300 : 500;
      assert(i<g_nMats);
      ierr = MatCreateSeqAIJ(PETSC_COMM_SELF,sz,sz,cm,PETSC_NULL,&g_matarr[i]);
      CHKERRQ(ierr);
      ierr = MatZeroEntries( g_matarr[i] ); CHKERRQ(ierr);
    }
    scall = MAT_INITIAL_MATRIX;
  } else {
    assert(0); // no refactoring!
    // put PETSc matrices back (extra) and then destroy them
    for (i=0; i<n_local_true; i++) {
      assert(i<g_nMats);
      Mat mymat=osm->pmat[i]; osm->pmat[i]=g_matarr[i]; g_matarr[i] = mymat;
      ierr = PromISGetSize( osm->is[i], &pm ); CHKERRQ(ierr);
      ierr = PromISGetSize( g_CRISs[i], &cm );       CHKERRQ(ierr);	
      if(cm+pm > max_sz) max_sz = cm+pm;
      ierr = MatSetOption( mymat, MAT_STRUCTURALLY_SYMMETRIC );  CHKERRQ(ierr);
    }
    /* 
       Destroy the blocks from the previous iteration
    */
    if (pc->flag == DIFFERENT_NONZERO_PATTERN) {
      ierr = MatDestroyMatrices(osm->n_local_true,&osm->pmat);CHKERRQ(ierr);
      scall = MAT_INITIAL_MATRIX;
    }
  }

  /* extract out the submatrices */
  ierr = MatGetSubMatrices(pc->pmat,n_local_true,osm->is,osm->is,scall,&osm->pmat);
  CHKERRQ(ierr);
  // put new matrices in (and copy)
  PromTable<int> geq_leq(4*max_sz/3+1);
  int bsz=PROM_BSZ,pbuffer[PROM_BSZ], *perm = pbuffer; 
  if( max_sz > bsz ) { 
    bsz = max_sz;
    ierr = PetscMalloc( bsz*sizeof(int), &perm ); CHKERRQ(ierr);
  }
  for( i = 0 ; i < n_local_true; i++ ){
    assert(i<g_nMats);
    int cnnz,nnz,*crids,*pids,j,k,creq;   
    const PetscScalar *vv; const PetscInt *ids;
    Mat pmat_bs = osm->pmat[i], mymat = g_matarr[i];
    ierr = PromISGetSize( osm->is[i], &pm );  CHKERRQ(ierr); // first cr eq
    PromIS cris = g_CRISs[i]; 
    ierr = PromISGetSize( cris, &cm ); CHKERRQ(ierr);
    // copy primal part
    for( j = nnz = 0 ; j < pm ; j++ ){
      ierr = MatGetRow( pmat_bs, j, &sz, &ids, &vv ); CHKERRQ(ierr);
      if(sz==0)PetscPrintf(PETSC_COMM_SELF,"[0]%s ERROR: zero row %d/%d\n",
			   __FUNCT__,j,pm);
      // add into 1,1 block (same indices)
      ierr = MatSetValues( mymat, 1, &j, sz, ids, vv, INSERT_VALUES ); 
      ierr = MatRestoreRow( pmat_bs, j, &sz, &ids, &vv ); CHKERRQ(ierr);
      nnz += sz;
    }
    // CC
    ierr = PromISGetIndices( cris, &crids ); CHKERRQ(ierr);
    ierr = PromISGetIndices( osm->is[i], &pids ); CHKERRQ(ierr);
    for(j=0 ; j<pm ; j++){int ad=geq_leq.add(j+1,pids[j]+1);assert(ad==1);}
    for(j=cnnz=0, creq=pm ; j < cm ; j++, creq++ ){
      const int eq = crids[j]; 
      ierr = MatGetRow( g_C, eq, &sz, &ids, &vv );  CHKERRQ(ierr);
      if( sz > bsz ) {
	if( perm != pbuffer ) PetscFree( perm );
	bsz = sz;
	ierr = PetscMalloc( bsz*sizeof(int), &perm ); CHKERRQ(ierr);
      }
      // get local ids
      for( k = 0 ; k < sz ; k++ ){
	perm[k] = geq_leq.find( ids[k]+1 ) - 1; 
	assert(perm[k] >= 0);
      } 
      // add in C and C'
      ierr = MatSetValues( mymat, 1, &creq, sz, perm, vv, INSERT_VALUES ); 
      CHKERRQ(ierr);
      ierr = MatSetValues( mymat, sz, perm, 1, &creq, vv, INSERT_VALUES ); 
      CHKERRQ(ierr);
      cnnz += 2*sz;
      ierr = MatRestoreRow( g_C, eq, &sz, &ids, &vv ); CHKERRQ(ierr);
    }
    ierr = PromISRestoreIndices( cris, &crids ); CHKERRQ(ierr);
    ierr = PromISRestoreIndices( osm->is[i], &pids ); CHKERRQ(ierr);
    // assemble
    ierr = MatAssemblyBegin( mymat, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
    ierr = MatAssemblyEnd( mymat, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
    IS ris, cis; 
    { 
      // get ordering (perm) 
      int j,n,k,r,*pp,ndf,eq; 
      ierr = MatOrdering_ND( pmat_bs, "nd", &ris, &cis ); CHKERRQ(ierr);
      ierr = PromISGetIndices( ris, &pp ); CHKERRQ(ierr);
      ierr = PromISGetSize( ris, &n );  CHKERRQ(ierr);
      ierr = MatGetBlockSize( pmat_bs, &ndf ); CHKERRQ(ierr);
      // get perm
      for(j=r=0;j<n;j++){for(k=0,eq=pp[j]*ndf;k<ndf;k++,eq++,r++)perm[r]=eq;}
      assert(r==pm);
      for(j=0;j<cm;j++,r++)perm[r]=r; // keep CRs the same
      assert(r==pm+cm);
      ierr = PromISRestoreIndices( ris, &pp ); CHKERRQ(ierr); 
      ierr = PromISDestroy( ris ); CHKERRQ(ierr); 
      ierr = PromISDestroy( cis ); CHKERRQ(ierr); 
    }
    // insert perm (hack), and setup 
    subksp = osm->ksp[i];  MatInfo info; 
    ierr = KSPGetPC( subksp, &subpc );         CHKERRQ(ierr);
    assert(subpc->setupcalled == 0);
    if( pm > 0 ) {
      PromISCreateGeneral( PETSC_COMM_SELF, pm+cm, perm, &cis );
      ierr = ISSetPermutation(cis); CHKERRQ(ierr);
      PromISCreateGeneral( PETSC_COMM_SELF, pm+cm, perm, &ris );
      ierr = ISSetPermutation(ris); CHKERRQ(ierr);
    }
    if( use_lu_subpc ) {// LU fill
      //
      double nz = (double)(nnz+cnnz)/double(pm+cm+1); 
      double fill = 1.0 + 1.5 * pow( double(pm), 0.7 ) / (nz+1.);
      ierr = PCLUSetFill( subpc, fill ); 	         CHKERRQ(ierr);  //
      // clone PETSc's pc setup
      PC_LU *dir = (PC_LU*)subpc->data;      assert(dir->row == NULL);
       if( pm > 0 ){
	dir->row = ris; dir->col = cis;      
	PetscLogObjectParent(subpc,dir->row);
	PetscLogObjectParent(subpc,dir->col);
	//
	ierr = MatLUFactorSymbolic(mymat,dir->row,dir->col,&dir->info,&dir->fact);
	CHKERRQ(ierr);
	ierr = MatGetInfo( dir->fact, MAT_LOCAL, &info ); CHKERRQ(ierr);
	dir->actualfill = info.fill_ratio_needed;
	PetscLogObjectParent(subpc,dir->fact); 
	ierr = MatLUFactorNumeric( mymat, &dir->info,&dir->fact ); CHKERRQ(ierr);
	subpc->setupcalled = 2;
      }
    }  
    else { // ILU
      ierr = PCILUSetAllowDiagonalFill( subpc ); CHKERRQ(ierr);
      if( 0 ){
	ierr = PromISDestroy( cis ); CHKERRQ(ierr);
	ierr = PromISDestroy( ris ); CHKERRQ(ierr);
	ierr = PCILUSetUseDropTolerance( subpc, 1.e-10, 1.e-1, 2*nnz/pm+1 );
	CHKERRQ( ierr );
	ierr = PCSetOperators( subpc, mymat, mymat, SAME_NONZERO_PATTERN );
	CHKERRQ( ierr );   
	//ierr = PCSetVector( subpc, osm->x[i] ); CHKERRQ(ierr);
	ierr = PCSetUp( subpc );  CHKERRQ(ierr);
      }
      else{
	int fill = (pm>1000) ? 1 : (pm>100) ? 2 : 3; 
	ierr = PCILUSetLevels( subpc, fill );	         CHKERRQ(ierr);
	ierr = PCILUSetFill( subpc, (double)(fill) + 1. );  CHKERRQ(ierr);
	PC_ILU *ilu = (PC_ILU*)subpc->data;
	assert(ilu->row == NULL);
	if( pm > 0 ) {
	  ilu->row = ris; ilu->col = cis;      
	  PetscLogObjectParent(subpc,ilu->row); 
	  PetscLogObjectParent(subpc,ilu->col);
	  ierr = MatILUFactorSymbolic(mymat,ilu->row,ilu->col,&ilu->info,
				      &ilu->fact);
	  CHKERRQ(ierr); 
	  ierr = MatLUFactorNumeric( mymat, &ilu->info,&ilu->fact ); CHKERRQ(ierr);
	  subpc->setupcalled = 2;
	}
      }
      assert(subpc->setupcalled == 2 || cm==0);
    }
    // cache old matrix and put new in -- ready to solve
    assert(i<g_nMats);
    osm->pmat[i] = mymat; g_matarr[i] = pmat_bs;
  }
  if( perm != pbuffer ) PetscFree( perm );
  /* Return control to the user so that the submatrices can be modified (e.g., to apply
     different boundary conditions for the submatrices than for the global problem) */
  //ierr = PCModifySubMatrices(pc,osm->n_local,osm->is,osm->is,osm->pmat,pc->modifysubmatricesP);CHKERRQ(ierr);
  /* loop over subdomains putting them into local sles */
  const char *pprefix;
  ierr = PetscObjectGetOptionsPrefix((PetscObject)pc->pmat,&pprefix);CHKERRQ(ierr);
  for (i=0; i<n_local_true; i++) {
    ierr = PetscObjectSetOptionsPrefix((PetscObject)osm->pmat[i],pprefix);CHKERRQ(ierr);
    PetscLogObjectParent(pc,osm->pmat[i]); // 'SAME_PRECONDITIONER' as we have just set it up
    ierr = KSPSetOperators(osm->ksp[i],osm->pmat[i],osm->pmat[i],SAME_PRECONDITIONER);CHKERRQ(ierr);
  }
  PetscFunctionReturn(0); // ???
}
#else // petra
#endif

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_ASM::Apply_private"
/* PromPCKKT_ASM::Apply_private ****************************************
 *
 *  - apply, additive Schwarz locally
 * 
 *  INPUT: 
 *   - XX: RHS
 *   - YY: answer
 *   - mult: do multiplicative
 */     
int PromPCKKT_ASM::Apply_private( const PromCRVector * const XX,
				  PromCRVector * const YY ) const
{
  assert(this!=NULL); assert(dpc_ != NULL && dpc_->pc_ != NULL); //
  PETSc_ASM *osm = (PETSc_ASM*)dpc_->pc_->data;
  int i,n_local = osm->n_local,n_local_true = osm->n_local_true,ierr;
  PetscScalar zero = 0.0; 
  ScatterMode forward = SCATTER_FORWARD,reverse = SCATTER_REVERSE;
  VecScatter* dScatters = (VecScatter*)d_scatters_; //

  PetscFunctionBegin;
  Vec xx = XX->vec_, yy = YY->vec_; //
  Vec yp = YY->p_->vec_, xp = XX->p_->vec_; assert(xp!=NULL); //
  /*
       Support for limiting the restriction or interpolation to only local 
     subdomain values (leaving the other values 0). 
  */
  if (!(osm->type & PC_ASM_RESTRICT)) {
    forward = SCATTER_FORWARD_LOCAL;
    /* have to zero the work RHS since scatter may leave some slots empty */
    for (i=0; i<n_local; i++) {
      ierr = VecSet(osm->x[i],zero);CHKERRQ(ierr);
    }
  }
  if (!(osm->type & PC_ASM_INTERPOLATE)) {
    reverse = SCATTER_REVERSE_LOCAL;
  }
  for (i=0; i<n_local; i++) {
    ierr = VecScatterBegin(xx,osm->x[i],INSERT_VALUES,forward,osm->scat[i]);CHKERRQ(ierr);
    ierr = VecScatterBegin(xp,osm->x[i],INSERT_VALUES,forward,dScatters[i]);CHKERRQ(ierr);//
  }
  ierr = VecSet(yy,zero);CHKERRQ(ierr);
  ierr = VecSet(yp,zero);CHKERRQ(ierr);

  /* do the local solves */
  for (i=0; i<n_local_true; i++) {
    ierr = VecScatterEnd(xx,osm->x[i],INSERT_VALUES,forward,osm->scat[i]);CHKERRQ(ierr);
    ierr = VecScatterEnd(xp,osm->x[i],INSERT_VALUES,forward,dScatters[i]);CHKERRQ(ierr);//
    //ierr = KSPSetSolution( osm->ksp[i], osm->y[i] ); CHKERRQ(ierr);
    //ierr = KSPSetRhs( osm->ksp[i], osm->x[i] );       CHKERRQ(ierr);
    //ierr = KSPSolve( osm->ksp[i], osm->x[i], osm->y[i] );                  CHKERRQ(ierr); 
    ierr = KSPSolve( osm->ksp[i], osm->x[i], osm->y[i] );CHKERRQ(ierr); 
    ierr = VecScatterBegin(osm->y[i],yy,ADD_VALUES,reverse,osm->scat[i]);CHKERRQ(ierr);
    ierr = VecScatterBegin(osm->y[i],yp,ADD_VALUES,reverse,dScatters[i]);CHKERRQ(ierr);//
  }

  /* handle the rest of the scatters that do not have local solves */
  for (i=n_local_true; i<n_local; i++) {
    ierr = VecScatterEnd(xx,osm->x[i],INSERT_VALUES,forward,osm->scat[i]);CHKERRQ(ierr);
    ierr = VecScatterEnd(xp,osm->x[i],INSERT_VALUES,forward,dScatters[i]);CHKERRQ(ierr);//
    ierr = VecScatterBegin(osm->y[i],yy,ADD_VALUES,reverse,osm->scat[i]);CHKERRQ(ierr);
    ierr = VecScatterBegin(osm->y[i],yp,ADD_VALUES,reverse,dScatters[i]);CHKERRQ(ierr);//
  }
  for (i=0; i<n_local; i++) {
    ierr = VecScatterEnd(osm->y[i],yy,ADD_VALUES,reverse,osm->scat[i]);CHKERRQ(ierr);
    ierr = VecScatterEnd(osm->y[i],yp,ADD_VALUES,reverse,dScatters[i]);CHKERRQ(ierr);//
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_ASM::SetNewOperator"
/* PromPCKKT_ASM::SetNewOperator ****************************************
 *
 *  INPUT: 
 *   - AA: primal part
 *   - CC: 
 */     
int PromPCKKT_ASM::SetNewOperator( const PromCRMatrix *const KK ) 
{
  assert( this != NULL ); 
  if( dpc_ != NULL ) { delete dpc_; dpc_ = NULL; }
  return 0; 
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_ASM::SetOperator"
/* PromPCKKT_ASM::SetOperator ****************************************
 *
 *  INPUT: 
 *   - AA: primal part
 *   - CC: 
 */     
int PromPCKKT_ASM::SetOperator( const PromCRMatrix *const KK ) 
{
  int ierr; assert( this != NULL );
  if( dpc_ != NULL ) {
    const PromMatrix *const AA = KK;
    ierr = dpc_->SetOperator( AA ); CHKERRQ(ierr);
  }
  return 0; 
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_ASM::SetNewPVec"
/* PromPCKKT_ASM::SetNewPVec ****************************************
 *
 */    
int PromPCKKT_ASM::SetNewPVec( const PromMap &map )
{
  if( work_ != NULL ) {
    // work_
    if( work_->p_ != NULL ) delete work_->p_;
    work_->p_ = new PromVector( map ); 
  }
  if( r_ != NULL ) {
    // r_
    if( r_->p_ != NULL ) delete r_->p_;
    r_->p_ = new PromVector( map ); 
  }

  return 0; 
} 

/************************************************************************
 *  PromPCKKT_Shell
 ************************************************************************/

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_Shell::getPCString"
const char * PromPCKKT_Shell::getPCString() const 
{
#if defined(PROM_USE_PETSC)
  PC pc = dpc_->pc_;  int  nlocal, first; KSP *subsles; PetscTruth islu;
  // set blocks up
  PCASMGetSubKSP( pc, &nlocal, &first, &subsles );
  KSPGetPC( subsles[0], &pc );   	 
  PetscTypeCompare((PetscObject)pc,PCLU,&islu); 
  if( islu ) return "KKT-LU"; 
  else return "KKT-ILU"; 
#else
  return "KKT-(I)LU"; 
#endif
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_Shell::CreatePrimAggs"
/* PromPCKKT_Shell::CreatePrimAggs ****************************************
 *
 *  - create agg of al locals 'PrimIS_'
 *
 *  INPUT: 
 *   - AA: primal matrix  
 */  
int PromPCKKT_Shell::CreatePrimAggs( const PromMatrix* const AA, bool ) 
{
  assert( this != NULL );  
  int ierr,ii,geq,ncols,*colids;
  PetscFunctionBegin;
  // top grid or processor block ILU with overlap on ghosts
  if( AA->getm() > 0 ) { // put all local in the IS, plus ghosts
    if( AA->getm() == AA->getM() ) { // top grid - simple
      PrimIS_ = make_is_w_all( AA );  assert(PrimIS_ != NULL);
    }
    else { // overlapping block ilu
      PromTable<int> ghost_lid(517);   assert(AA->map_.isConstNDF()); 
      const int ndf = AA->map_.constNDFSize(), my0 = AA->map_.my0Eq();
      const int ncrs = C_->getm(), next0 = my0 + AA->getm();
      for(ii=0,geq=C_->map_.my0Eq();ii<ncrs;ii++,geq++){
	ierr = C_->GetRow(geq, &ncols, &colids ); CHKERRQ(ierr);
	for( int jj=0 ; jj < ncols ; /*void*/ ) {
	  const int geq0 = (colids[jj]/ndf)*ndf; 
	  if( geq0 < my0 || geq0 >= next0 ) {
	    for(int kk=0;kk<ndf;kk++) ghost_lid.add(geq0 + kk + 1);
	  }
	  // inc 'jj' to next geq0
	  while( jj < ncols && colids[jj] < geq0+ndf ) jj++;
	}
	ierr = C_->RestoreRow( geq, &ncols, &colids );CHKERRQ(ierr);
      }
      // create IS
      assert(PrimIS_==NULL);
      int buff__[4096],*buff,neq=AA->getm(),nn = neq + ghost_lid.getCount();
      ierr = PetscMalloc(sizeof(PromIS), &PrimIS_); CHKERRQ(ierr);
      if(nn<4096) buff = buff__;
      else PetscMalloc( sizeof(int)*nn, &buff );
      for(ii=0, geq=my0; ii<neq ; ii++, geq++ ) buff[ii] = geq; // all locals
      while( !ghost_lid.isEmpty() ) buff[ii++] = ghost_lid.removeHead() - 1;
      assert(ii == nn);
      PromISCreateGeneral( PETSC_COMM_SELF, nn, buff, &PrimIS_[0] );
      ierr = ISSort( PrimIS_[0] );                             CHKERRQ(ierr);
      if(buff != buff__) PetscFree(buff);
    }
  }
  else{
    assert(PrimIS_ == NULL);
    ierr = PetscMalloc(sizeof(PromIS), &PrimIS_); CHKERRQ(ierr);
    ierr = PromISCreateGeneral(PETSC_COMM_SELF, 0, NULL, &PrimIS_[0]);
    CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_Shell::make_is_w_all"
/* PromPCKKT_Shell::make_is_w_all ****************************************
 *
 *  INPUT: 
 *    - mat: matrix
 */     
PromIS *PromPCKKT_Shell::make_is_w_all( const PromMatrix *const mat ) 
{
  PromIS *out = NULL;
  PetscFunctionBegin; 
  // create 'CRIS'
  PetscMalloc( sizeof(PromIS), &out ); assert(out != NULL);
  if( mat->getm() > 0 ) { // put all local in the IS
    int buff__[4096],*buff,nn=mat->getm(),eq=mat->map_.my0Eq(); 
    if(nn<4096) buff = buff__;
    else PetscMalloc( sizeof(int)*nn, &buff );
    for( int ii=0 ; ii<nn ; ii++, eq++ ) buff[ii] = eq;
    PromISCreateGeneral( PETSC_COMM_SELF, nn, buff, &out[0]);
    if(buff != buff__) PetscFree(buff);
  }
  else {
    PromISCreateGeneral( PETSC_COMM_SELF, 0, NULL, &out[0]);
  }
  PetscFunctionReturn(out);
  //PetscStackPop; 
  //return (out);
}

/************************************************************************
 *  PromPCKKT_Seg
 ************************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_Seg::Create"
/* PromPCKKT_Seg::Create ****************************************
 *
 *  - create 'dpc_' 
 *
 *  INPUT: 
 *   - opts:
 *   - perf:
 *   - AA: primal part
 *
 *   SIDE EFFECTS:
 *
 *   RETURN:
 *     - PETSc error code 
 */
int PromPCKKT_Seg::Create( const PromOptions &opts, 
			   PromPerfMonitor &perf,
			   const PromMatrix *const AA,
			   const PromCRVector *const XX, bool ilu /*=FALSE*/ ) 
{
  assert(this!=NULL); assert(dpc_==NULL); assert(CDinvCt_!=NULL); assert(!ilu);
  int ierr;

  PetscFunctionBegin; // dual part
  dpc_ = new PromPC( opts, perf, AA->MPIComm() );
  ierr = dpc_->SetType( PROMPCASM, XX->p_ );  CHKERRQ(ierr);
  ierr = dpc_->SetOperator( CDinvCt_ );  CHKERRQ(ierr);
#if defined(PROM_USE_PETSC)
  int nlocal, first_local; KSP *subsles;
  ierr = PCASMSetOverlap( dpc_->pc_, 0 );            CHKERRQ(ierr);
  ierr = PCASMSetType( dpc_->pc_, PC_ASM_RESTRICT ); CHKERRQ(ierr);
  ierr = PCSetUp( dpc_->pc_ ); CHKERRQ(ierr);
  ierr = PCASMGetSubKSP(dpc_->pc_, &nlocal, &first_local,&subsles);  
  CHKERRQ(ierr); assert(nlocal==1);
  PC pc;
  ierr = KSPGetPC( subsles[0], &pc );   	 CHKERRQ(ierr);
  ierr = PCSetType( pc, PCLU );			 CHKERRQ(ierr);
  ierr = PCLUSetMatOrdering( pc, MATORDERING_ND ); CHKERRQ(ierr);
  int sz = CDinvCt_->getm(); 
  if( sz > 0 ) {
    double nnz = CDinvCt_->getLocalNNZ()/sz; 
#if defined(PARCH_ascired)
    double fill = (sz > 0) ? (nnz + 8.0*sqrt((double)sz))/nnz : 1.0;
#else
    double fill = (sz > 0) ? (nnz + 5.0*pow((double)sz,0.65))/nnz : 1.0;
#endif
    if( fill < 1.0 ) fill = 1.0;
    ierr = PCLUSetFill( pc, fill ); 	        CHKERRQ(ierr);
  }
  else {
    ierr = PCLUSetFill( pc, 1.0 ); 	        CHKERRQ(ierr);
  }
  // real setup (factors)
  ierr = PCSetUpOnBlocks( dpc_->pc_ );          CHKERRQ(ierr);
  dpc_->setup_called_ = 2; // all setup -- like 'CreateASM'
#endif
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_Seg::SetOperator"
/* PromPCKKT_Seg::SetOperator ****************************************
 *
 *  INPUT: 
 *   - AA: primal part
 *   - CCt: dual PC operator
 *
 *   SIDE EFFECTS:
 *
 *   RETURN:
 *     - PETSc error code 
 */     
int PromPCKKT_Seg::SetOperator( const PromCRMatrix *const KK ) 
{
  int ierr; assert( this != NULL );

  CDinvCt_ = KK->CDinvCt_;
  C_ = KK->C_;
  //A_ = dynamic_cast<const PromMatrix*>(KK); assert(A_!=NULL);
  ierr = dpc_->SetOperator( CDinvCt_ );   CHKERRQ(ierr);

  return 0; 
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_Seg::SetNewOperator"
/* PromPCKKT_Seg::SetNewOperator ****************************************
 *
 *  INPUT: 
 *   - AA: primal part
 *   - CCt: dual PC operator
 *
 *   SIDE EFFECTS:
 *
 *   RETURN:
 *     - PETSc error code 
 */     
int PromPCKKT_Seg::SetNewOperator( const PromCRMatrix *const KK ) 
{
  int ierr; assert( this != NULL );

  CDinvCt_ = KK->CDinvCt_;
  C_ = KK->C_;
  //A_ = dynamic_cast<const PromMatrix*>(KK); assert(A_!=NULL);
  ierr = dpc_->SetNewOperator( CDinvCt_ );   CHKERRQ(ierr);
  
  return 0; 
}

/************************************************************************
 * PromProjKKTSolver - for (few) large CR eqs. w/ uzawa (ie, not all-at-once)
 ************************************************************************/

#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::SetProjC"
/* PromProjKKTSolver::SetProjC *****************************************
 *
 *  - setup projection operator, construct projection data 
 *
 *   INPUT:
 *     - C:
 *     - CRCache: (destroyed)
 *
 *   SIDE EFFECTS:
 * 
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::SetProjC( const PromMatrix * const C,
				 PromTable<int> *CReq_projCReq )
{
  assert(this!=NULL);   assert(projC_ == NULL);
  int ierr,ii,ssz,tt,jj,geq,*sbuff,*pb,**iarrs,*sizes,ncols,*colids,*tp; 
  const int np=C->map_.comm_.np(),mype=C->map_.comm_.mype(); 
  const int my0 = C->map_.my0Eq(); const int tag = C->map_.comm_.getNewTag(); 
  ii = sizeof(double);  ssz = sizeof(MPI_Request); double *vv, **varrs;
  const int fisz = ii/sizeof(int), nCRs = C->getM(), nlocCRs = C->getm();
  MPI_Status status;   MPI_Comm comm = C->map_.comm_.MPIComm();
  MPI_Request reqs[PROM_BSZ];

  PetscFunctionBegin;
  projC_ = C; assert( nCRs > 0 ); assert(ipiv_ == NULL);assert(CCt_inv_==NULL);
  ierr = PetscMalloc( nCRs*sizeof(int), &ipiv_);             CHKERRQ(ierr);
  memset((void*)ipiv_, 99, nCRs*sizeof(int));
  ierr = PetscMalloc( nCRs*nCRs*sizeof(double), &CCt_inv_);  CHKERRQ(ierr);
  memset((void*)CCt_inv_, 99,  nCRs*nCRs*sizeof(double));
  // set CR --> projCR map
  if( nlocCRs ) {
    assert(CReq_projCReq->getCount() == nlocCRs); 
    ierr = PetscMalloc(nlocCRs*sizeof(int),&projLEq_geq_);CHKERRQ(ierr);
    memset((void*)projLEq_geq_, 99,  nlocCRs*sizeof(int));
    for(ii = 0 ; ii < nlocCRs ; ii++) projLEq_geq_[ii] = -1;
    TablePos tpos = CReq_projCReq->getHeadPosition(); 
    while( tpos ) {
      int creq, projeq = CReq_projCReq->getNext( tpos, &creq ) - 1; creq--;
      int id = projeq - my0; assert(id >= 0 && id < nlocCRs);
      projLEq_geq_[id] = creq;
    }
    for(ii = 0 ; ii < nlocCRs ; ii++) assert(projLEq_geq_[ii] != -1);
  }
  assert(workCR_ == NULL);
  workCR_ = new PromVector( projC_->map_ );

  // count sends
  for( ii = 0, ssz = 0, geq = my0 ; ii < nlocCRs ; ii++, geq++ ) {
    ierr = C->GetRow( geq, &ncols );   CHKERRQ(ierr);
    ssz += 2 + ncols; if( ssz%fisz != 0 ) ssz++; 
    ssz += ncols * fisz;
    ierr = C->RestoreRow( geq, &ncols ); CHKERRQ(ierr);
  }
  ierr = PetscMalloc( (ssz+1)*sizeof(int), &sbuff ); CHKERRQ(ierr);
  // write sends, save arrays (can be zero), store CRs
  ierr = PetscMalloc( nCRs*sizeof(int*), &iarrs );     CHKERRQ(ierr);
  ierr = PetscMalloc( nCRs*sizeof(double*), &varrs );  CHKERRQ(ierr);
  ierr = PetscMalloc( nCRs*sizeof(int), &sizes );      CHKERRQ(ierr);
  for(ii=0; ii<nCRs; ii++){sizes[ii]=-1; iarrs[ii]=NULL; varrs[ii]=NULL;}
  for( pb = sbuff, ii = 0, geq = my0 ; ii < nlocCRs ; ii++, geq++ ){
    ierr = C->GetRow( geq, &ncols, &colids, &vv );   CHKERRQ(ierr);
    *pb++ = geq; *pb++ = ncols;
    if( ncols > 0 ) {
      assert(varrs[geq] == NULL);
      ierr = PetscMalloc( ncols*sizeof(int), &iarrs[geq]); CHKERRQ(ierr);
      ierr = PetscMalloc( ncols*sizeof(double), &varrs[geq]); CHKERRQ(ierr);
      tp = iarrs[geq];
      for( jj = 0 ; jj < ncols ; jj++ ) *pb++ = tp[jj] = colids[jj];
      if( (pb-sbuff)%fisz != 0 ) *pb++ = -3;
      double *fp = (double*)pb, *tfp = varrs[geq];
      for( jj=0;jj<ncols;jj++) *fp++ = tfp[jj] = vv[jj];
      pb = (int*)fp;
    }
    assert(sizes[geq] == -1);
    sizes[geq] = ncols; // cache
    ierr = C->RestoreRow( geq, &ncols, &colids, &vv );   CHKERRQ(ierr);
  }
  if(pb-sbuff != ssz)SETERRQ2(1,"pb-sbuff != ssz %d %d",pb-sbuff,ssz);
  // send(tag) <eq, len, len*<ids>, [-3,] len*<vals> >
  for( ii = 0 ; ii < np ; ii++ ){
    if( ii != mype ) {
      MPI_Isend( sbuff, ssz, MPI_INT, ii, tag, comm, &reqs[ii] );
    }
  }
  // recv(tag) <eq, len, len*<ids>, [-3,] len*<vals> >
  for( ii = 0 ; ii < np - 1 ; ii++ ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, comm, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    if( tt == 0 ) {
      MPI_Recv( NULL, 0, MPI_INT, status.MPI_SOURCE, tag, comm, &status );
    }
    else {
      int *buf; 
      ierr = PetscMalloc(tt*sizeof(int),&buf); pb = buf; CHKERRQ(ierr);
      // recv(tag)  *< eq, len, len*<ids>, [-3,] len*<vals> > 
      MPI_Recv( buf, tt, MPI_INT, status.MPI_SOURCE, tag, comm, &status );
      while( pb-buf < tt ) {
	geq = *pb++; ncols = *pb++; assert(geq<nCRs);
	if( ncols > 0 ) {
	  assert(varrs[geq] == NULL);
	  ierr = PetscMalloc( ncols*sizeof(int), &iarrs[geq]); CHKERRQ(ierr);
	  ierr = PetscMalloc( ncols*sizeof(double),&varrs[geq]); CHKERRQ(ierr);
	  tp = iarrs[geq];
	  for( jj = 0 ; jj < ncols ; jj++ ) tp[jj] = *pb++;
	  if( (size_t)pb%sizeof(double) != 0 ) {int t = *pb++; assert(t==-3); }
	  double *fp = (double*)pb, *tfp = varrs[geq];
	  for( jj=0;jj<ncols;jj++) tfp[jj] = *fp++;
	  pb = (int*)fp;
	}
     	assert(sizes[geq] == -1);
	sizes[geq] = ncols;// cache
      }
      PetscFree(buf); assert(pb-buf == tt);
    }
  }
  // local (redundent) form (CC')^-1
  for( geq = 0 ; geq < nCRs ; geq++){
    int *ip1 = iarrs[geq], nz1 = sizes[geq]; double *vv1 = varrs[geq];
    assert(nz1 != -1);
    // 
    for( int geq2 = geq ; geq2 < nCRs ; geq2++ ){
      int *ip2=iarrs[geq2],nz2=sizes[geq2]; double *vv2 = varrs[geq2], sum;
      SparseDot( nz1, ip1, vv1, nz2, ip2, vv2, sum );
      int id = geq + nCRs*geq2, idt = geq2 + nCRs*geq; 
      CCt_inv_[id] = CCt_inv_[idt] = sum;
    }
  }

  if( 0 ){ // debug
    FILE *file; char path[128]; int mype; int i,j,k;
    MPI_Comm_rank( PETSC_COMM_WORLD, &mype ); 
    sprintf( path, "./projCCt_%d",mype);
    file = fopen( path, "w");
    for(k=i=0;i<nCRs;i++)for(j=0;j<nCRs;j++,k++)
      fprintf(file,"%d %d %le\n",i+1,j+1,CCt_inv_[k]);
    fclose(file);
  }

  // invert
  dgetrf_( &nCRs, &nCRs, CCt_inv_, &nCRs, ipiv_, &ierr );
  
  // wait for sends
  for( ii = 0 ; ii < np ; ii++ ){
    if( ii != mype ) MPI_Wait( &reqs[ii], &status ); 
  }
  for( ii = 0 ; ii < nCRs ; ii++ ){ 
    assert(sizes[ii] != -1);
    if( sizes[ii] > 0 ) {
      PetscFree(iarrs[ii]); PetscFree(varrs[ii]);
    }
  }
  PetscFree(sbuff); PetscFree(iarrs); PetscFree(varrs); PetscFree(sizes);
  
  PetscFunctionReturn(0);  
}
#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::SparseDot"
/* PromProjKKTSolver::SparseDot **************************************
 *
 *   INPUT:
 *     - nz1:
 *     - ip1[]:
 *     - vv1[]:
 *     - nz2:
 *     - ip2[]:
 *     - vv2[]:
 *     - dot (out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::SparseDot(const int nz1, const int ip1[],
				 const double vv1[],
				 const int nz2, const int ip2[], 
				 const double vv2[], double &dot)
{
  // sparse dot product
  int k1 = 0, k2 = 0; dot = 0.0;
  while( k1 < nz1 && k2 < nz2 ) {
    while( k1 < nz1 && ip1[k1] < ip2[k2] ) k1++;
    if(k1 == nz1) break;
    while( k2 < nz2 && ip2[k2] < ip1[k1] ) k2++;
    if(k2 == nz2) break;
    if( ip1[k1] == ip2[k2] ){
      dot += vv1[k1] * vv2[k2];
      k1++; k2++;
    }
  }
  return 0;
}


#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::MapAllCRsToProjCRs"
/* PromProjKKTSolver::MapAllCRsToProjCRs **************************************
 *
 *   INPUT:
 *     - XX[nCRs]:
 *     - YY[nProjCRs]: (out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::MapAllCRsToProjCRs( const PromVector * const GG,
					   PromVector * const YY ) const
{
  assert(this!=NULL); 
  int ierr,ii,projEq;  double *gg;  
  const int nlocprojcr = YY->getn(), my0pcr = YY->map_.my0Eq();
  const int myCR0 = GG->map_.my0Eq(), nlocCR = GG->getn();
  assert(nlocprojcr == projC_->map_.nlocalEq());
  ierr = GG->GetArray( &gg ); CHKERRQ(ierr); 
  for( ii = 0, projEq = my0pcr ; ii < nlocprojcr ; ii++, projEq++ ) {
    int creq = projLEq_geq_[ii], lcreq = creq-myCR0;
    ierr = YY->SetValues( 1, &projEq, &gg[lcreq], INSERT_VALUES );
    CHKERRQ(ierr); assert(lcreq>=0 && lcreq<nlocCR);
  }
  ierr = GG->RestoreArray( &gg );    CHKERRQ(ierr);

  return 0;
}
 
#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::Apply_Ct_CCInv"
/* PromProjKKTSolver::Apply_Ct_CCInv *****************************************
 *
 *   - Apply: YY <-  C' * (CC')^-1 @ GG
 *   - used to get conforming initial guess from "gap" 'g'
 *
 *   INPUT:
 *     - GG[gross nCR]: 
 *     - YY[nPrimal]: (out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::Apply_Ct_CCInv( const PromVector * const GG,
				PromVector * const YY ) const
{
  int ierr;    assert(this!=NULL); 
  assert(projC_!=NULL && workCR_!=NULL);
  
  // map GG to 'workCR_'
  ierr = MapAllCRsToProjCRs( GG, workCR_ );                 CHKERRQ(ierr);
  // all to all: workCR_ <-  (CC')^-1 workCR_
  ierr = ApplyCCInv( workCR_ );                       CHKERRQ(ierr);
  // Y <- C' workCR_ 
  ierr = projC_->MultTranspose( workCR_, YY ); CHKERRQ(ierr);

#if defined(PETSC_USE_BOPT_g)
  double beta,beta2;
  ierr = GG->Norm2(&beta); CHKERRQ(ierr); 
  ierr = YY->Norm2(&beta2); CHKERRQ(ierr); 
  if( (beta==0.0) != (beta2==0.0) ) SETERRQ2(1,"|x|=%e |y|=%e",beta,beta2);
#endif

  return 0;
}
 
#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::Apply_Ct_add"
/* PromProjKKTSolver::Apply_Ct_add *****************************************
 *
 *   - Apply: YY <-  YY + C' * @ GG
 *   - used to form residual
 *
 *   INPUT:
 *     - GG[nCRs]: 
 *     - XX[nPrimal]: 
 *     - YY[nPrimal]: (in/out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::Apply_Ct_add( const PromVector * const GG,
				     PromVector * const XX,
				     PromVector * const YY ) const
{
  int ierr;    assert(this!=NULL); 
  assert(projC_!=NULL && workCR_!=NULL);
  
  // map GG to 'workCR_'
  ierr = MapAllCRsToProjCRs( GG, workCR_ );                       CHKERRQ(ierr);
  // Y <- Y + C' workCR_ 
  ierr = projC_->MultTransposeAdd( workCR_, XX, YY ); CHKERRQ(ierr);
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::MapProjCRsToAllCRs"
/* PromProjKKTSolver::MapProjCRsToAllCRs **********************************
 *
 *   INPUT:
 *     - XX[nProjCRs]: 
 *     - YY[nCRs]: (in/out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::MapProjCRsToAllCRs( const PromVector * const XX,
					   PromVector * const YY ) const
{
  assert(this!=NULL); 
  double *rr;     int ii,ierr;
  const int nlocpcr = XX->getn();
  assert(nlocpcr == projC_->map_.nlocalEq());

  ierr = XX->GetArray( &rr ); CHKERRQ(ierr); 
  for( ii = 0 ; ii < nlocpcr ; ii++ ) {
    int creq = projLEq_geq_[ii];
    ierr = YY->SetValues( 1, &creq, &rr[ii], INSERT_VALUES ); CHKERRQ(ierr);
  }
  ierr = XX->RestoreArray( &rr );    CHKERRQ(ierr);
  
  return 0;
}
#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::Apply_CCInv_C"
/* PromProjKKTSolver::Apply_CCInv_C *****************************************
 *
 *   - Apply: YY <- @ (CC')^-1 * C * RR
 *   - used to recover lagrange multiplier
 *
 *   INPUT:
 *     - RR[nPrimal]: residual
 *     - YY[gross nCR]: lagrange multipliers for projected CRs (in/out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::Apply_CCInv_C( const PromVector * const RR,
				      PromVector * const YY ) const
{
  int ierr;    assert(this!=NULL);   
  assert(projC_!=NULL && workCR_!=NULL);
  
  // workCR_ <- C RR
  ierr = projC_->Mult( RR, workCR_ ); CHKERRQ(ierr);
  // all to all: workCR_ <-  (CC')^-1 workCR_
  ierr = ApplyCCInv( workCR_ );                 CHKERRQ(ierr);
  // map 'workCR_' to YY
  ierr = MapProjCRsToAllCRs ( workCR_, YY );     CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::ApplyProj"
/* PromProjKKTSolver::ApplyProj *****************************************
 *
 *   - Apply: YY <-  [WW +] ( I - C' * (CC')^-1 * C ) [A] XX
 *
 *   INPUT:
 *     - XX[nPrimal]: 
 *     - YY[nPrimal]: (out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::ApplyProj( const PromVector * const XX,
				  PromVector * const YY ) const
{
  int ierr;    assert(this!=NULL); 
  assert(projC_!=NULL && workCR_!=NULL);

  // workCR_ <- C  X
  ierr = projC_->Mult( XX, workCR_ );             CHKERRQ(ierr);
  // all to all: workCR_ <-  (CC')^-1 workCR_
  ierr = ApplyCCInv( workCR_ );                   CHKERRQ(ierr);
  // Y <- C' workCR_ 
  ierr = projC_->MultTranspose( workCR_, YY );  CHKERRQ(ierr);
  // Y <- X - Y = X - C' (CC')^-1 C X = (I - C' (CC')^-1 C) X = (I - P) X
  ierr = YY->PromVector::AYPX( -1.0, XX );   CHKERRQ(ierr);

#if defined(PETSC_USE_BOPT_g)
  double beta; 
  if( !isConforming( YY, beta ) ) {
    if( beta > 1.e-4 ){ 
      FILE *file; char path[128]; int mype, npe;double *vv; int ii,nn;
      MPI_Comm_size( PETSC_COMM_WORLD, &npe ); 
      MPI_Comm_rank( PETSC_COMM_WORLD, &mype ); 
      // Ct_CCtInv_CX
      if( npe == 1 ) sprintf(path, "./Ct_CCtInv_CX");
      else sprintf( path, "./Ct_CCtInv_CX_%d",mype);
      file = fopen( path, "w");
      ierr = YY->GetArray( &vv ); CHKERRQ(ierr); 
      nn=YY->map_.nlocalEq();
      for(ii=0;ii<nn;ii++)fprintf(file,"%le\n",vv[ii]); 
      ierr = YY->RestoreArray( &vv ); CHKERRQ(ierr);       fclose(file);
      // progC
      if( npe == 1 ) sprintf( path, "./progC");
      else sprintf( path, "./progC_%d", mype);
      file = fopen( path, "w");
      ierr = PromContext::WriteMat_private(projC_,file,0,0,FALSE);fclose(file);
      // XX
      if( npe == 1 ) sprintf(path, "./X");
      else sprintf( path, "./X_%d",mype);
      file = fopen( path, "w");
      ierr = XX->GetArray( &vv ); CHKERRQ(ierr); 
      nn=XX->map_.nlocalEq();
      for(ii=0;ii<nn;ii++)fprintf(file,"%le\n",vv[ii]); 
      ierr = XX->RestoreArray( &vv ); CHKERRQ(ierr);       fclose(file);
      // CX
      projC_->Mult( XX, workCR_ );
      if( npe == 1 ) sprintf(path, "./CX");
      else sprintf( path, "./CX_%d",mype);
      file = fopen( path, "w");
      ierr = workCR_->GetArray( &vv ); CHKERRQ(ierr); 
      nn=workCR_->map_.nlocalEq();
      for(ii=0;ii<nn;ii++)fprintf(file,"%le\n",vv[ii]); 
      ierr = workCR_->RestoreArray( &vv ); CHKERRQ(ierr);       fclose(file);
      // CCtInv_CX
      ApplyCCInv( workCR_ ); 
      if( npe == 1 ) sprintf(path, "./CCtInv_CX");
      else sprintf( path, "./CCtInv_CX_%d",mype);
      file = fopen( path, "w");
      ierr = workCR_->GetArray( &vv ); CHKERRQ(ierr); 
      nn=workCR_->map_.nlocalEq();
	for(ii=0;ii<nn;ii++)fprintf(file,"%le\n",vv[ii]); 
      ierr = workCR_->RestoreArray( &vv ); CHKERRQ(ierr);       fclose(file);
      SETERRQ1(1,"bad projection %e",beta); 
    }
    else PetscPrintf(PETSC_COMM_SELF,
		     "[?]%s warning poor projection %g\n",__FUNCT__,beta);
  }
#endif
  return 0;
}
#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::isConforming"
/* PromProjKKTSolver::isConforming *****************************************
 *
 *   - 
 *
 *   INPUT:
 *     - XX[nPrimal]: 
 *     - beta: (out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - is the solver conforming
 */
bool PromProjKKTSolver::isConforming( const PromVector * const XX, 
				      double &beta ) const
{
  double beta2; int ierr;
  beta = 0.0; // out
  ierr = XX->Norm2( &beta2 ); if(ierr) return FALSE;
  if( beta2 == 0.0 ) return TRUE; 
  ierr = projC_->Mult( XX, workCR_ ); if(ierr) return FALSE;
  ierr = workCR_->Norm2( &beta ); if(ierr) return FALSE;
  beta = beta/beta2;
  return (beta < 1.e-9);
}

#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::ApplyCCInv"  
/* PromProjKKTSolver::ApplyCCInv *****************************************
 *
 *   - Apply: GG <-  (CC')^-1 GG
 *
 *   INPUT:
 *     - GG: (in/out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::ApplyCCInv( PromVector * const GG ) const
{
  int ierr; assert(this!=NULL);

  assert( projC_ != NULL );
  int i1=1,nloc=GG->getn(),nCRs=GG->getN(),my0=projC_->map_.my0Eq(); 
  MPI_Comm comm = projC_->map_.comm_.MPIComm();
  double CRarr[1024]; assert(nCRs<1024);
  // get 'CRarr' -- global!
  {
    double *yy; 
    const int *displs = projC_->map_.proc_geq(), np = projC_->map_.comm_.np(); 
    int ii, recvcounts[PROM_BSZ];
    ierr = GG->GetArray( &yy ); CHKERRQ(ierr);
    for(ii=0;ii<np;ii++)recvcounts[ii] = displs[ii+1] - displs[ii];
    MPI_Allgatherv(yy,nloc,MPI_DOUBLE,CRarr,recvcounts,(int*)displs,MPI_DOUBLE,comm);
    ierr = GG->RestoreArray( &yy );    CHKERRQ(ierr);
  }
  // apply (CC')^-1
  dgetrs_( "N", &nCRs, &i1, CCt_inv_, &nCRs, ipiv_, CRarr, &nCRs, &ierr );
  
  {  // put back int GG
    int ips[1024]; assert(nloc<1024);
    for( int ii = 0, jj = my0 ; ii < nloc ; ii++, jj++ ) ips[ii] = jj;
    ierr = GG->SetValues(nloc, ips, &CRarr[my0], INSERT_VALUES);CHKERRQ(ierr);
  }
  CHKERRQ(ierr);
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::PreSolve"
/* PromProjKKTSolver::PreSolve *****************************************
 *
 *   - prep for solve
 *
 *   INPUT:
 *     - A:
 *     - gg[nCD]: 
 *     - fp[nPrime]: f <- (I - P)(f - A x_0) (in/out)
 *     - x_0[nPrime]: C' * (CC')^-1 gg (out)
 *     - work[nPrime]: temp
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::PreSolve( const PromMatrix * const A,
				 const PromVector* const gg, 
				 PromVector * const fp, 
				 PromVector * const x_0 ) const
{
  int ierr; assert(this!=NULL);
  
  if( ksp_mult_work_ == NULL ) ksp_mult_work_ = new PromVector( A->map_ );
  PromVector *work = ksp_mult_work_;

  // x_0 <- C' * (CC')^-1 g
  ierr = Apply_Ct_CCInv( gg, x_0 );     CHKERRQ(ierr);
  // residual: r <- f - A x_0  
  ierr = A->PromMatrix::Mult( x_0, work );  CHKERRQ(ierr);
  ierr = work->PromVector::AYPX( -1.0, fp );  CHKERRQ(ierr);
  // f_p <- (I - P)(f - A x_0)
  ierr = ApplyProj( work, fp ); CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::PostSolve"
/* PromProjKKTSolver::PostSolve *****************************************
 *
 *   - unwind solution: x <- x_0 + (I - P)x
 *
 *   INPUT:
 *     - x_0: conforming part of solution, used as work space
 *     - xx: (in/out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::PostSolve( const PromVector * const x_0,
				  PromVector * const xx ) const
{
  int ierr; assert(this!=NULL);
  
  if( ksp_mult_work_ == NULL ) ksp_mult_work_ = new PromVector( xx );
  PromVector *work = ksp_mult_work_;

#if defined(PETSC_USE_BOPT_g)
  double beta; 
  if( !isConforming( x_0, beta ) ) SETERRQ1(1,"|C *x_0|/|x_0| (!=0) =%e",beta);
#endif  
  // x <- x_0 + (I - P)x
  ierr = ApplyProj( xx, work );               CHKERRQ(ierr);
  ierr = xx->PromVector::WAXPY( 1.0, x_0, work );  CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromProjKKTSolver::ComputeMultiplier"
/* PromProjKKTSolver::ComputeMultiplier ************************************
 *
 *   - unwind solution
 *
 *   INPUT:
 *     - A:
 *     - Cother: ie, uzawa, rest of C
 *     - xx: (in/out)
 *     - pp: (in/out)
 *     - bb:
 *     - x_0: conforming part of solution, used as work space
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromProjKKTSolver::ComputeMultiplier( const PromMatrix * const A,
					  const PromMatrix * const otherC,
					  const PromVector * const xx, 
					  PromVector * const pp,
					  const PromVector * const bb ) const
{
  int ierr; assert(this!=NULL);
  
  if( ksp_mult_work_ == NULL ) ksp_mult_work_ = new PromVector( A->map_ );
  PromVector *work = ksp_mult_work_;
  
  // form residual
  ierr = A->PromMatrix::Mult( xx, work );                    CHKERRQ(ierr);
  ierr = otherC->MultTransposeAdd( pp, work, work );     CHKERRQ(ierr);
  ierr = work->PromVector::AYPX( -1.0, bb );                 CHKERRQ(ierr);
  // YY <- @ (CC')^-1 * C * RR
  ierr = Apply_CCInv_C( work, pp );  CHKERRQ(ierr);
  
  return 0;
}

/************************************************************************
 *  PromSolver support for Projection solver
 ************************************************************************/

#undef __FUNCT__  
#define __FUNCT__ "PromSolver::KSPMult"
/* PromSolver::KSPMult *****************************************
 *
 *   INPUT:
 *     - A:
 *     - XX: 
 *     - YY: (out)
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromSolver::KSPMult( const PromMatrix_base * const A,
			 const PromVector_base * const XX, 
			 PromVector_base * const YY ) const
{
  int ierr, pind = PromContext::prom_logs_[LOG_AUX1];  assert(this!=NULL); 
  if( !alloced_ ) {
    ierr = alias_->KSPMult( A, XX, YY ); CHKERRQ(ierr);
  }
  else {
    assert(alloced_);  assert(XX!=YY);
    // do projectiom mult 
    if( projector_ == NULL ) {
      ierr = A->Mult( XX, YY ); CHKERRQ(ierr);
    }
    else { 
      if( projector_->ksp_mult_work_ == NULL ) {
	projector_->ksp_mult_work_ = new PromVector( A->map_ );
      }
      const PromVector *xx;
      PromVector *yy, *work = projector_->ksp_mult_work_;
      xx = dynamic_cast<const PromVector*>(XX); assert(xx!=NULL);
      yy = dynamic_cast<PromVector*>(YY); assert(yy!=NULL);
      // yy <- (I - P) A (I - P) xx
      perf_mon_.EventBegin(pind);
      ierr = projector_->ApplyProj( xx, yy ); CHKERRQ(ierr);
      perf_mon_.EventEnd(pind);
      // doit
      ierr = A->Mult( yy, work ); CHKERRQ(ierr);
      // 
      perf_mon_.EventBegin(pind);
      ierr = projector_->ApplyProj( work, yy ); CHKERRQ(ierr);
      perf_mon_.EventEnd(pind);
    }
  }
  return 0;
}

/************************************************************************
 * Apply methods for KKT-MG
 ************************************************************************/

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_Shell::Apply"
/* PromPCKKT_Shell::Apply ****************************************
 *
 *  - Apply
 *
 *  *  INPUT: 
 */  
int PromPCKKT_Shell::Apply(PromPC_base *const /*pc*/, 
			   const PromCRMatrix *const AA,
			   const PromCRVector*const bb, PromCRVector*const xx,
			   const bool zeroguess/*=TRUE*/ )
{
  int ierr; assert(this!=NULL); 
  // just apply kkt solver -- 'pc' should be empty
  if( !zeroguess ) {
    if( work_ == NULL ) {
      work_ = new PromCRVector( AA->map_, &C_->map_ );
    } 
    if( r_ == NULL ) {
      r_ = new PromCRVector( AA->map_, &C_->map_ );
    }
    // residual: r <- f - A x_0  
    ierr = AA->Mult( xx, r_ );  CHKERRQ(ierr);
    ierr = r_->AYPX( -1.0, bb );  CHKERRQ(ierr);
    //ierr = work_->Set( 0.0 );  CHKERRQ(ierr);
    ierr = Apply_private( r_, work_ ); CHKERRQ(ierr); /* x <- B*r */
    ierr = xx->AXPY( 1.0, work_ );CHKERRQ(ierr); /* x' <- x'+t */
  }
  else {
    ierr = Apply_private( bb, xx ); CHKERRQ(ierr); /* x <- B*r */
  }
  return 0; 
}  

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_ASM::Apply"
/* PromPCKKT_ASM::Apply ****************************************
 *
 *  - Additive, and multiplicative, Vanka smoother
 * 
 *  INPUT: 
 *   - pc: pc for primal
 *   - AA: primal part
 *   - bb: RHS
 *   - xx: answer
 *   - mult: do multiplicative
 */     
int PromPCKKT_ASM::Apply( PromPC_base *const pc, 
			  const PromCRMatrix *const AA,
			  const PromCRVector * const bb,
			  PromCRVector * const xx,
			  const bool zeroguess ) 
{
  assert(this!=NULL); assert(AA->map_.isConstNDF()); 
  const int pre = pc->takeNonZeroGuess(); // hack for mult (G-S)
  int ierr; assert(zeroguess || pre != 0);

  // get operators 
  assert(AA != NULL && C_ != NULL);
  if( work_ == NULL ) work_ = new PromCRVector( AA->map_, &C_->map_ );
  if(r_ == NULL) r_ = new PromCRVector( AA->map_, &C_->map_ );
  
  // apply
  if( pre == 0 ) { // additive
    // apply primal -- if( zeroguess || pre != 0 ) {
    ierr = pc->Apply_private(bb, xx, zeroguess); CHKERRQ(ierr); /* t <- B*r*/
    if( nMask_ > 0 ) {
      ierr = xx->PromVector::SetValues(nMask_,mask_,zeros_,INSERT_VALUES);
      CHKERRQ(ierr);
    }
    // apply dual
    ierr = Apply_private( bb, work_ ); CHKERRQ(ierr); /* t <- B*r */
#if defined(PETSC_USE_BOPT_g)
    if(zeroguess){
      for(int i=0;i<nMask_;i++)assert(zeros_[i]==0.0);
      double beta; xx->Dot(work_,&beta); 
      if(beta!=0.0)SETERRQ1(1,"x't != 0.0 (%g)",beta);
    }
#endif
    ierr = xx->AXPY( 1.0, work_ );     CHKERRQ(ierr); /* x' <- x'+t */
  }
  else if( pre == 1 ||1){ // front mult. -- cheap w/ zero guess
    // apply dual
    if( zeroguess ) {
      ierr = Apply_private( bb, xx );      CHKERRQ(ierr); /* t <- B*r */
    }
    else{
      ierr = AA->Mult( xx, r_ );        CHKERRQ(ierr); /* r <= b - Ax */ 
      ierr = r_->AYPX( -1.0, bb );      CHKERRQ(ierr);
      //
      ierr = Apply_private( r_, work_ );   CHKERRQ(ierr); /* t <- B*r */
      ierr = xx->AXPY( 1.0, work_ );          CHKERRQ(ierr); /* x <- x + t */
    }
    // apply primal -- have init guess
    ierr = C_->MultTranspose( xx->p_, r_ );  CHKERRQ(ierr);/* r <- C'*p */
    ierr = r_->PromVector::AYPX( -1.0, bb ); CHKERRQ(ierr);/* r <- b - C'*p*/
    //
    ierr = pc->Apply_private(r_, xx, FALSE); CHKERRQ(ierr); /* x <- x + B*r */
  }
  else{ // back mult. 
    // apply primal -- put primal residual in 'r_'
    if( zeroguess ) {
      ierr = pc->Apply_private( bb, xx, TRUE );CHKERRQ(ierr); /* x <- B*r */
    } 
    else{ // !zero guess
      ierr = C_->MultTranspose( xx->p_, r_ );  CHKERRQ(ierr);/* r <- C'*p */
      ierr = r_->PromVector::AYPX( -1.0, bb ); CHKERRQ(ierr);/* r <- b - C'*p*/
      //
      ierr = pc->Apply_private(r_, xx, FALSE); CHKERRQ(ierr);/* x <- x + B*r */
    }
    // dual part of r
    ierr = AA->Mult( xx, r_ );        CHKERRQ(ierr); /* r <= b - Ax */ 
    ierr = r_->AYPX( -1.0, bb );      CHKERRQ(ierr);
    // apply dual
    ierr = Apply_private( r_, work_ );     CHKERRQ(ierr); /* t <- B*r */
    ierr = xx->AXPY( 1.0, work_ );         CHKERRQ(ierr); /* x' <- x'+t */
  }

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromPCKKT_Seg::Apply"
/* PromPCKKT_Seg::Apply ****************************************
 *
 *  - Breass smoother
 *
 *  INPUT: 
 *   - pc:
 *   - AA:
 *   - bb:
 *   - xx
 *   - zeroguess
 *
 *   SIDE EFFECTS:
 *
 *   RETURN:
 *     - PETSc error code 
 */     
int PromPCKKT_Seg::Apply(PromPC_base *const pc, const PromCRMatrix *const AA,
			 const PromCRVector*const bb, PromCRVector*const xx,
			 const bool zeroguess/*TRUE*/ )
{
  assert(this!=NULL); assert(dpc_ != NULL); assert(C_ != NULL);
  int ierr; 
  const PromMatrix *A = dynamic_cast<const PromMatrix*>(AA); assert(A!=NULL);
  
  // buffers
  if( work_ == NULL ) {
    work_ = new PromCRVector( AA->map_, &C_->map_ ); assert(r_ == NULL);
    r_ = new PromCRVector( AA->map_, &C_->map_ );
  }

  // primal part x <- x + B( f - Ax - Cp ) 
  if( !zeroguess ) {
    ierr = C_->MultTranspose( xx->p_, r_ ); CHKERRQ(ierr);/* r <- Cp */
    //
    if( pc->takeNonZeroGuess() != 0 ){
      //ierr=work_->PromVector::Copy(xx); CHKERRQ(ierr); // copy x_0 to 'work'
      ierr = r_->PromVector::AYPX(-1.0,bb); CHKERRQ(ierr);/* r <- b - Cp */
      ierr = pc->Apply_private(r_,xx,FALSE);CHKERRQ(ierr);/* x <- x + B*r */
    }
    else {
      // residual: r <- f - A x_0 - C x_0
      ierr = AA->PromMatrix::MultAdd( xx, r_, r_ ); CHKERRQ(ierr);      
      ierr = r_->PromVector::AYPX( -1.0, bb );  CHKERRQ(ierr);
      //
      ierr = pc->Apply_private( r_, work_ );  CHKERRQ(ierr); /* x <- x + B*r */
      ierr = xx->PromVector::AXPY(1.0,work_); CHKERRQ(ierr); /* x' <- x' + t */
    }
  }
  else {
    ierr = pc->Apply_private( bb, xx );CHKERRQ(ierr);    /* x <- x + B*r */
    xx->p_->Set(0.0);
  }

  // dual part
  ierr = C_->Mult( xx, r_->p_ );       CHKERRQ(ierr);
  ierr = r_->p_->AXPY( -1.0, bb->p_ ); CHKERRQ(ierr);    /* r <- Cp - g */
  ierr = dpc_->Apply_private( r_->p_, work_->p_, TRUE ); /* tp <- Q(Cp-g) */
  CHKERRQ(ierr); 
  ierr = xx->p_->AXPY( 1.0, work_->p_); CHKERRQ(ierr); /* p <- p + tp */

  // symm -- r <- C' * tp
  ierr = C_->MultTranspose( work_->p_, r_ ); CHKERRQ(ierr);
  // x <- x - Q r 
  ierr = pc->Apply_private( r_, work_ ); CHKERRQ(ierr);  /* x <- -B*r */
  ierr = xx->PromVector::AXPY( -1.0, work_ ); CHKERRQ(ierr);
  
  return 0; 
}
